from __future__ import annotations

from functools import wraps
from typing import Any, Callable, Optional, Type, TypeVar, cast, overload

from expression.core import Error, Ok, Result

_TSource = TypeVar("_TSource")
_TError = TypeVar("_TError", bound=Exception)
_TError_ = TypeVar("_TError_", bound=Exception)


@overload
def catch(
    exception: Type[_TError_],
) -> Callable[
    [Callable[..., _TSource | Result[_TSource, _TError]]],
    Callable[..., Result[_TSource, _TError | _TError_]],
]:
    ...


@overload
def catch(
    f: Callable[..., _TSource], *, exception: Type[_TError]
) -> Callable[..., Result[_TSource, _TError]]:
    ...


def catch(  # type: ignore
    f: Optional[Callable[..., _TSource]] = None, *, exception: Type[_TError]
) -> Callable[
    [Callable[..., _TSource]],
    Callable[..., Result[_TSource, _TError]] | Result[_TSource, _TError],
]:
    def decorator(
        fn: Callable[..., _TSource]
    ) -> Callable[..., Result[_TSource, _TError]]:
        @wraps(fn)
        def wrapper(*args: Any, **kwargs: Any) -> Result[_TSource, _TError]:
            try:
                out = fn(*args, **kwargs)
            except exception as exn:
                return Error(cast(_TError, exn))
            else:
                if isinstance(out, Result):
                    return cast(Result[_TSource, _TError], out)

                return Ok(out)

        return wrapper

    if f is not None:
        return decorator(f)

    return decorator


__all__ = ["catch"]



"""The aiotools (async) module.

The aio (asyncio) module contains asynchronous utility functions for
working with async / await.

The module is inspired by the F# `Async` module, but builds on top of
Python async / await instead of providing an asynchronous IO mechanism
by itself.
"""
import asyncio
from asyncio import Future
from typing import Any, Awaitable, Callable, Optional, TypeVar

from expression.system import CancellationToken, OperationCanceledError

_TSource = TypeVar("_TSource")

Continuation = Callable[[_TSource], None]
Callbacks = Callable[
    [
        Continuation[_TSource],
        Continuation[Exception],
        Continuation[OperationCanceledError],
    ],
    None,
]


def from_continuations(callback: Callbacks[_TSource]) -> Awaitable[_TSource]:
    """Creates an asynchronous computation that captures the current
    success, exception and cancellation continuations. The callback must
    eventually call exactly one of the given continuations.

    Args:
        callback: The function that accepts the current success,
            exception, and cancellation continuations.

    Returns:
        An asynchronous computation that provides the callback with the
        current continuations.
    """
    future: "Future[_TSource]" = asyncio.Future()

    def done(value: _TSource) -> None:
        future.set_result(value)

    def error(err: Exception) -> None:
        future.set_exception(err)

    def cancel(_: OperationCanceledError) -> None:
        future.cancel()

    callback(done, error, cancel)
    return future


def start(
    computation: Awaitable[Any], token: Optional[CancellationToken] = None
) -> None:
    """Starts the asynchronous computation in the event loop. Do not
    await its result.

    If no cancellation token is provided then the default cancellation
    token is used.
    """

    async def runner() -> Any:
        return await computation

    task = asyncio.create_task(runner())

    def cb():
        task.cancel()

    if token:
        token.register(cb)
    return None


def start_immediate(
    computation: Awaitable[Any], token: Optional[CancellationToken] = None
) -> None:
    """Runs an asynchronous computation, starting immediately on the
    current operating system thread."""

    async def runner() -> Any:
        return await computation

    task = asyncio.create_task(runner())

    def cb() -> None:
        task.cancel()

    if token:
        token.register(cb)
    return None


def run_synchronously(computation: Awaitable[_TSource]) -> _TSource:
    """Runs the asynchronous computation and await its result."""

    async def runner() -> _TSource:
        return await computation

    return asyncio.run(runner())


async def singleton(value: _TSource) -> _TSource:
    """Async function that returns a single value."""
    return value


async def sleep(msecs: int) -> None:
    """Creates an asynchronous computation that will sleep for the given
    time. This is scheduled using a System.Threading.Timer object. The
    operation will not block operating system threads for the duration
    of the wait."""
    await asyncio.sleep(msecs / 1000.0)


async def empty() -> None:
    """Async no-op"""


def from_result(result: _TSource) -> Awaitable[_TSource]:
    """Creates a async operation that's completed successfully with the
    specified result."""

    async def from_result(result: _TSource) -> _TSource:
        """Async return value"""
        return result

    return from_result(result)


__all__ = [
    "Continuation",
    "empty",
    "from_continuations",
    "singleton",
    "sleep",
    "start",
    "start_immediate",
    "run_synchronously",
]



from typing import Any, Callable, Tuple, TypeVar

_A = TypeVar("_A")
_B = TypeVar("_B")
_TSource = TypeVar("_TSource")
_TResult = TypeVar("_TResult")


def identity(value: _A) -> _A:
    """Identity function.

    Returns value given as argument.
    """
    return value


def starid(*value: Any) -> Tuple[Any, ...]:
    return value


def flip(fn: Callable[[_A, _B], _TResult]) -> Callable[[_B, _A], _TResult]:
    """Flips the arguments for a function taking two arguments.

    Example:
        >>> fn(a, b) == flip(fn)(b, a)
    """

    def _flip(b: _B, a: _A) -> Any:
        return fn(a, b)

    return _flip


def snd(value: Tuple[Any, _TSource]) -> _TSource:
    """Return second argument of the tuple."""

    _, b = value
    return b


def fst(value: Tuple[_TSource, Any]) -> _TSource:
    """Return first argument of the tuple."""

    a, _ = value
    return a


__all__ = ["fst", "identity", "starid", "flip", "snd"]



"""Option module.

Contains a collection of static methods (functions) for operating on
options. All functions takes the source as the last curried argument,
i.e all functions returns a function that takes the source sequence as
the only argument.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    Iterator,
    List,
    Optional,
    TypeVar,
    cast,
)

from typing_extensions import TypeGuard

from .curry import curry_flip
from .error import EffectError
from .pipe import PipeMixin
from .typing import GenericValidator, ModelField, SupportsValidation

if TYPE_CHECKING:
    from ..collections.seq import Seq

_TSource = TypeVar("_TSource")
_TResult = TypeVar("_TResult")

_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")


def _validate(value: Any, field: ModelField) -> Option[Any]:
    if isinstance(value, Option):
        return cast(Option[Any], value)

    if isinstance(value, Dict) and not value:
        return Nothing

    if field.sub_fields:
        sub_field = field.sub_fields[0]
        value, error = sub_field.validate(value, {}, loc="Option")
        if error:
            raise ValueError(str(error))

    return Some(cast(Any, value))


class Option(
    Iterable[_TSource],
    PipeMixin,
    SupportsValidation["Option[_TSource]"],
    ABC,
):
    """Option abstract base class."""

    __validators__: List[GenericValidator[Option[_TSource]]] = [_validate]

    @abstractmethod
    def default_value(self, value: _TSource) -> _TSource:
        """Get with default value.

        Gets the value of the option if the option is Some, otherwise
        returns the specified default value.
        """
        raise NotImplementedError

    @abstractmethod
    def default_with(self, getter: Callable[[], _TSource]) -> _TSource:
        """Get with default value lazily.

        Gets the value of the option if the option is Some, otherwise
        returns the value produced by the getter
        """
        raise NotImplementedError

    @abstractmethod
    def map(self, mapper: Callable[[_TSource], _TResult]) -> Option[_TResult]:
        raise NotImplementedError

    @abstractmethod
    def map2(
        self, mapper: Callable[[_TSource, _T2], _TResult], other: Option[_T2]
    ) -> Option[_TResult]:
        raise NotImplementedError

    @abstractmethod
    def bind(self, mapper: Callable[[_TSource], Option[_TResult]]) -> Option[_TResult]:
        """Bind option.

        Applies and returns the result of the mapper if the value is
        `Some`. If the value is `Nothing` then `Nothing` is returned.

        Args:
            mapper: A function that takes the value of type TSource from
                an option and transforms it into an option containing a
                value of type TResult.

        Returns:
            An option of the output type of the mapper.
        """
        raise NotImplementedError

    @abstractmethod
    def or_else(self, if_none: Option[_TSource]) -> Option[_TSource]:
        """Returns option if it is Some, otherwise returns `if_one`."""
        raise NotImplementedError

    @abstractmethod
    def or_else_with(self, if_none: Callable[[], Option[_TSource]]) -> Option[_TSource]:
        """Returns option if it is Some,
        otherwise evaluates the given function and returns the result."""
        raise NotImplementedError

    @abstractmethod
    def filter(self, predicate: Callable[[_TSource], bool]) -> Option[_TSource]:
        """Returns the input if the predicate evaluates to true,
        otherwise returns `Nothing`"""
        raise NotImplementedError

    @abstractmethod
    def to_list(self) -> List[_TSource]:
        raise NotImplementedError

    @abstractmethod
    def to_seq(self) -> Seq[_TSource]:
        raise NotImplementedError

    @abstractmethod
    def is_some(self) -> bool:
        """Returns true if the option is not Nothing."""
        raise NotImplementedError

    @abstractmethod
    def is_none(self) -> bool:
        """Returns true if the option is Nothing."""
        raise NotImplementedError

    @classmethod
    def of_obj(cls, value: _TSource) -> Option[_TSource]:
        """Convert object to an option."""
        return of_optional(value)

    @classmethod
    def of_optional(cls, value: Optional[_TSource]) -> Option[_TSource]:
        """Convert optional value to an option."""
        return of_optional(value)

    @abstractmethod
    def dict(self) -> _TSource | Dict[str, Any]:
        """Returns a json string representation of the option."""
        raise NotImplementedError

    @property
    @abstractmethod
    def value(self) -> _TSource:
        """Returns the value wrapped by the option.

        A `ValueError` is raised if the option is `Nothing`.
        """
        raise NotImplementedError

    @abstractmethod
    def __eq__(self, o: Any) -> bool:
        raise NotImplementedError

    @abstractmethod
    def __iter__(self) -> Generator[_TSource, _TSource, _TSource]:
        raise NotImplementedError

    @abstractmethod
    def __lt__(self, other: Any) -> bool:
        raise NotImplementedError

    def __repr__(self) -> str:
        return self.__str__()

    @classmethod
    def __get_validators__(cls) -> Iterator[GenericValidator[Option[_TSource]]]:
        yield from cls.__validators__

    @abstractmethod
    def __hash__(self) -> int:
        raise NotImplementedError


class Some(Option[_TSource]):
    """The Some option case class."""

    __match_args__ = ("value",)

    def __init__(self, value: _TSource) -> None:
        self._value = value

    def default_value(self, value: _TSource) -> _TSource:
        """Gets the value of the option if the option is Some, otherwise
        returns the specified default value.
        """
        return self._value

    def default_with(self, getter: Callable[[], _TSource]) -> _TSource:
        """Get with default value.

        Gets the value of the option if the option is Some, otherwise
        returns the value produced by the getter
        """
        return self._value

    def is_some(self) -> bool:
        """Returns `True`."""
        return True

    def is_none(self) -> bool:
        """Returns `False`."""
        return False

    def map(self, mapper: Callable[[_TSource], _TResult]) -> Option[_TResult]:
        return Some(mapper(self._value))

    def map2(
        self, mapper: Callable[[_TSource, _T2], _TResult], other: Option[_T2]
    ) -> Option[_TResult]:
        if isinstance(other, Some):
            return Some(mapper(self._value, other.value))
        return Nothing

    def bind(self, mapper: Callable[[_TSource], Option[_TResult]]) -> Option[_TResult]:
        """Bind option.

        Applies and returns the result of the mapper if the value is
        `Some`. If the value is `Nothing` then `Nothing` is returned.

        Args:
            mapper: A function that takes the value of type TSource from
                an option and transforms it into an option containing a
                value of type TResult.

        Returns:
            An option of the output type of the mapper.
        """
        return mapper(self._value)

    def or_else(self, if_none: Option[_TSource]) -> Option[_TSource]:
        """Returns `self`."""
        return self

    def or_else_with(self, if_none: Callable[[], Option[_TSource]]) -> Option[_TSource]:
        """Returns `self`."""
        return self

    def filter(self, predicate: Callable[[_TSource], bool]) -> Option[_TSource]:
        """Returns the input if the predicate evaluates to true,
        otherwise returns `Nothing`"""
        return self if predicate(self._value) else Nothing

    def to_list(self) -> List[_TSource]:
        return [self._value]

    def to_seq(self) -> Seq[_TSource]:
        # deferred import to avoid circular dependencies
        from expression.collections.seq import Seq

        return Seq.of(self._value)

    def dict(self) -> _TSource:
        attr = getattr(self._value, "dict", None) or getattr(self._value, "dict", None)
        if attr and callable(attr):
            value = attr()
        else:
            value = self._value

        return value

    @property
    def value(self) -> _TSource:
        """Returns the value wrapped by the option.

        A `ValueError` is raised if the option is `Nothing`.
        """
        return self._value

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, Some):
            return self._value < other._value  # type: ignore
        return False

    def __eq__(self, o: Any) -> bool:
        if isinstance(o, Some):
            return self._value == o._value  # type: ignore
        return False

    def __iter__(self) -> Generator[_TSource, _TSource, _TSource]:
        return (yield self._value)

    def __str__(self) -> str:
        return f"Some {self._value}"

    def __hash__(self) -> int:
        return hash(self._value)


class Nothing_(Option[_TSource], EffectError):
    """The None option case class.

    Do not use. Use the singleton `Nothing` instead. Since Nothing is a
    singleton it can be tested e.g using `is`:
        >>> if xs is Nothing:
        ...     return True
    """

    def default_value(self, value: _TSource) -> _TSource:
        """Gets the value of the option if the option is Some, otherwise
        returns the specified default value.
        """
        return value

    def default_with(self, getter: Callable[[], _TSource]) -> _TSource:
        """Get with default value.

        Gets the value of the option if the option is Some, otherwise
        returns the value produced by the getter
        """
        return getter()

    def is_some(self) -> bool:
        """Returns `False`."""
        return False

    def is_none(self) -> bool:
        """Returns `True`."""
        return True

    def map(self, mapper: Callable[[_TSource], _TResult]) -> Option[_TResult]:
        return Nothing

    def map2(
        self, mapper: Callable[[_TSource, _T2], _TResult], other: Option[_T2]
    ) -> Option[_TResult]:
        return Nothing

    def bind(self, mapper: Callable[[_TSource], Option[_TResult]]) -> Option[_TResult]:
        """Bind option.

        Applies and returns the result of the mapper if the value is
        `Some`. If the value is `Nothing` then `Nothing` is returned.

        Args:
            mapper: A function that takes the value of type TSource from
                an option and transforms it into an option containing a
                value of type TResult.

        Returns:
            An option of the output type of the mapper.
        """
        return Nothing

    def or_else(self, if_none: Option[_TSource]) -> Option[_TSource]:
        """Returns `if_none`."""
        return if_none

    def or_else_with(self, if_none: Callable[[], Option[_TSource]]) -> Option[_TSource]:
        """Evaluates `if_none` and returns the result."""
        return if_none()

    def filter(self, predicate: Callable[[_TSource], bool]) -> Option[_TSource]:
        return Nothing

    def to_list(self) -> List[_TSource]:
        return []

    def to_seq(self) -> Seq[_TSource]:
        # deferred import to avoid circular dependencies
        from expression.collections.seq import Seq

        return Seq()

    def dict(self) -> Dict[str, Any]:
        return {}  # Pydantic cannot handle None or other types than Optional

    @property
    def value(self) -> _TSource:
        """Returns the value wrapped by the option.

        A `ValueError` is raised if the option is `Nothing`.
        """

        raise ValueError("There is no value.")

    def __iter__(self) -> Generator[_TSource, _TSource, _TSource]:
        """Return iterator for the `Nothing` case.

        We basically want to return nothing, but we have to return
        something to signal fail.
        """

        raise Nothing
        while False:
            yield

    def __lt__(self, other: Any) -> bool:
        return True

    def __eq__(self, o: Any) -> bool:
        if o is Nothing:
            return True
        return False

    def __str__(self):
        return "Nothing"

    def __hash__(self) -> int:
        return 0


# The singleton None class. We use the name 'Nothing' here instead of `None` to
# avoid conflicts with the builtin `None` value in Python.
# Note to self: Must be of type `Nothing_` or pattern matching will not work.
Nothing: Nothing_[Any] = Nothing_()
"""Singleton `Nothing` object.

Since Nothing is a singleton it can be tested e.g using `is`:
    >>> if xs is Nothing:
    ...     return True
"""


@curry_flip(1)
def bind(
    option: Option[_TSource], mapper: Callable[[_TSource], Option[_TResult]]
) -> Option[_TResult]:
    """Bind option.

    Applies and returns the result of the mapper if the value is
    `Some`. If the value is `Nothing` then `Nothing` is returned.

    Args:
        mapper: A function that takes the value of type _TSource from
            an option and transforms it into an option containing a
            value of type TResult.

    Returns:
        A partially applied function that takes an option and returns an
        option of the output type of the mapper.
    """

    return option.bind(mapper)


@curry_flip(1)
def default_value(option: Option[_TSource], value: _TSource) -> _TSource:
    """Gets the value of the option if the option is Some, otherwise
    returns the specified default value.
    """

    return option.default_value(value)


def default_with(
    getter: Callable[[], _TSource]
) -> Callable[[Option[_TSource]], _TSource]:
    """Get with default value lazily.

    Gets the value of the option if the option is Some, otherwise
    returns the value produced by the getter
    """

    def _default_with(option: Option[_TSource]) -> _TSource:
        return option.default_with(getter)

    return _default_with


def is_none(option: Option[_TSource]) -> TypeGuard[Nothing_[_TSource]]:
    return option.is_none()


def is_some(option: Option[_TSource]) -> TypeGuard[Some[_TSource]]:
    return option.is_some()


@curry_flip(1)
def map(
    option: Option[_TSource], mapper: Callable[[_TSource], _TResult]
) -> Option[_TResult]:
    return option.map(mapper)


@curry_flip(2)
def map2(
    opt1: Option[_T1], opt2: Option[_T2], mapper: Callable[[_T1, _T2], _TResult]
) -> Option[_TResult]:
    return opt1.map2(mapper, opt2)


def or_else(
    option: Option[_TSource],
    if_none: Option[_TSource],
) -> Option[_TSource]:
    """Returns option if it is Some, otherwise returns `if_none`."""
    return option.or_else(if_none)


def to_list(option: Option[_TSource]) -> List[_TSource]:
    return option.to_list()


def to_seq(option: Option[_TSource]) -> Seq[_TSource]:
    return option.to_seq()


def of_optional(value: Optional[_TSource]) -> Option[_TSource]:
    """Convert an optional value to an option.

    Convert a value that could be `None` into an `Option` value. Same as
    `of_obj` but with typed values.

    Args:
        value: The input optional value.

    Return:
        The result option.
    """
    if value is None:
        return Nothing

    return Some(value)


def of_obj(value: Any) -> Option[Any]:
    """Convert object to an option.

    Convert a value that could be `None` into an `Option` value.

    Args:
        value: The input object.

    Return:
        The result option.
    """
    return of_optional(value)


def dict(value: Option[_TSource]) -> _TSource | Dict[Any, Any] | None:
    return value.dict()


def default_arg(value: Option[_TSource], default_value: _TSource) -> _TSource:
    """Specify default argument.

    Used to specify a default value for an optional argument in the
    implementation of a function. Same as `default_value`, but
    "uncurried" and with the arguments swapped.
    """
    return value.default_value(default_value)


__all__ = [
    "Option",
    "Some",
    "Nothing",
    "Nothing_",
    "bind",
    "default_arg",
    "default_value",
    "default_with",
    "map",
    "map2",
    "is_none",
    "is_some",
    "or_else",
    "to_list",
    "dict",
    "to_seq",
    "of_optional",
    "of_obj",
]



"""Sequence module.

Contains a collection of static methods (functions) for operating on
sequences. A sequence is a thin wrapper around `Iterable` so all
functions take (and return) Python iterables.

All functions takes the source as the last curried
argument, i.e all functions returns a function that takes the source
sequence as the only argument.

Example (functional style):
    >>> from expression.collections import seq
    >>> xs = seq.of_iterable([1, 2, 3])
    >>> ys = xs.pipe(
        seq.map(lambda x: x + 1),
        seq.filter(lambda x: x < 3)
    )

Example (fluent style):
    >>> from expression.collections import Seq
    >>> xs = Seq([1, 2, 3])
    >>> ys = xs.map(lambda x: x + 1).filter(lambda x: x < 3)
"""
from __future__ import annotations

import builtins
import functools
import itertools
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Iterator,
    Tuple,
    TypeVar,
    cast,
    overload,
)

from expression.core import (
    Option,
    PipeMixin,
    SupportsGreaterThan,
    SupportsLessThan,
    SupportsSum,
    curry_flip,
    identity,
    pipe,
)

if TYPE_CHECKING:
    from .block import Block

_TSource = TypeVar("_TSource")
_TResult = TypeVar("_TResult")
_TState = TypeVar("_TState")
_TSupportsGreaterThan = TypeVar("_TSupportsGreaterThan", bound=SupportsGreaterThan)
_TSupportsLessThan = TypeVar("_TSupportsLessThan", bound=SupportsLessThan)
_TSupportsSum = TypeVar("_TSupportsSum", bound=SupportsSum)
_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
_T3 = TypeVar("_T3")
_T4 = TypeVar("_T4")


class Seq(Iterable[_TSource], PipeMixin):
    """Sequence type.

    Contains instance methods for dot-chaining operators methods on
    sequences.

    Example:
        >>> xs = Seq([1, 2, 3])
        >>> ys = xs.map(lambda x: x + 1).filter(lambda x: x < 3)
    """

    __match_args__ = ("iterable",)

    def __init__(self, iterable: Iterable[_TSource] = ()) -> None:
        self._value = iterable

    @classmethod
    def of(cls, *args: _TSource) -> Seq[_TSource]:
        return cls(args)

    @classmethod
    def of_iterable(cls, source: Iterable[_TSource]) -> Seq[_TSource]:
        return cls(source)

    def append(self, *others: Iterable[_TSource]) -> Seq[_TSource]:
        """Wraps the two given enumerations as a single concatenated
        enumeration."""
        return Seq(concat(self._value, *others))

    def filter(self, predicate: Callable[[_TSource], bool]) -> Seq[_TSource]:
        return Seq(filter(predicate)(self))

    def choose(self, chooser: Callable[[_TSource], Option[_TResult]]) -> Seq[_TResult]:
        """Choose items from the sequence.

        Applies the given function to each element of the list. Returns
        the list comprised of the results x for each element where the
        function returns `Some(x)`.

        Args:
            chooser: The function to generate options from the elements.

        Returns:
            The list comprising the values selected from the chooser
            function.
        """

        xs = pipe(self, choose(chooser))
        return Seq(xs)

    def collect(self, mapping: Callable[[_TSource], "Seq[_TResult]"]) -> Seq[_TResult]:
        xs = pipe(self, collect(mapping))
        return Seq(xs)

    @staticmethod
    def delay(generator: Callable[[], Iterable[_TSource]]) -> Iterable[_TSource]:
        """Returns a sequence that is built from the given delayed specification of a
        sequence.

        The input function is evaluated each time an IEnumerator for the sequence
        is requested.

        Args:
            generator: The generating function for the sequence.
        """

        return delay(generator)

    @staticmethod
    def empty() -> Seq[_TSource]:
        """Returns empty sequence."""
        return Seq()

    def fold(
        self, folder: Callable[[_TState, _TSource], _TState], state: _TState
    ) -> _TState:
        """Applies a function to each element of the collection,
        threading an accumulator argument through the computation. If
        the input function is f and the elements are i0...iN then
        computes f (... (f s i0)...) iN

        Args:
            folder: A function that updates the state with each element
                from the sequence.
            state: The initial state.
        Returns:
            The state object after the folding function is applied to
            each element of the sequence.
        """
        return functools.reduce(folder, self, state)

    def head(self) -> _TSource:
        """Returns the first element of the sequence."""

        return head(self)

    def length(self) -> int:
        """Returns the length of the sequence."""
        return length(self)

    def map(self, mapper: Callable[[_TSource], _TResult]) -> Seq[_TResult]:
        """Map sequence.

        Builds a new collection whose elements are the results of
        applying the given function to each of the elements of the
        collection.

        Args:
            mapping: A function to transform items from the input sequence.

        Returns:
            The result sequence.
        """

        return Seq(pipe(self, map(mapper)))

    @overload
    def starmap(
        self: Seq[Tuple[_T1, _T2]], mapping: Callable[[_T1, _T2], _TResult]
    ) -> Seq[_TResult]:
        ...

    @overload
    def starmap(
        self: Seq[Tuple[_T1, _T2, _T3]],
        mapping: Callable[[_T1, _T2, _T3], _TResult],
    ) -> Seq[_TResult]:
        ...

    @overload
    def starmap(
        self: Seq[Tuple[_T1, _T2, _T3, _T4]],
        mapping: Callable[[_T1, _T2, _T3, _T4], _TResult],
    ) -> Seq[_TResult]:
        ...

    def starmap(self: Seq[Tuple[Any, ...]], mapping: Callable[..., Any]) -> Seq[Any]:
        """Starmap source sequence.

        Unpack arguments grouped as tuple elements. Builds a new collection
        whose elements are the results of applying the given function to the
        unpacked arguments to each of the elements of the collection.

        Args:
            mapping: A function to transform items from the input sequence.

        Returns:
            Partially applied map function.
        """

        return Seq(starmap(mapping)(self))

    def mapi(self, mapping: Callable[[int, _TSource], _TResult]) -> Seq[_TResult]:
        """Map list with index.

        Builds a new collection whose elements are the results of
        applying the given function to each of the elements of the
        collection. The integer index passed to the function indicates
        the index (from 0) of element being transformed.

        Args:
            mapping: The function to transform elements and their
                indices.

        Returns:
            The list of transformed elements.
        """
        return Seq(mapi(mapping)(self))

    @overload
    @staticmethod
    def range(stop: int) -> Iterable[int]:
        ...

    @overload
    @staticmethod
    def range(start: int, stop: int) -> Iterable[int]:
        ...

    @overload
    @staticmethod
    def range(start: int, stop: int, step: int) -> Iterable[int]:
        ...

    @staticmethod
    def range(*args: int, **kw: int) -> Iterable[int]:
        return range(*args, **kw)

    def scan(
        self, scanner: Callable[[_TState, _TSource], _TState], state: _TState
    ) -> Iterable[_TState]:
        """Like fold, but computes on-demand and returns the sequence of
        intermediary and final results.

        Args:
            scanner: A function that updates the state with each element
                from the sequence.
            state: The initial state.

        Returns:
            The resulting sequence of computed states.
        """
        return Seq(itertools.accumulate(self._value, scanner, initial=state))

    def skip(self, count: int) -> Seq[_TSource]:
        """Returns a sequence that skips N elements of the underlying
        sequence and then yields the remaining elements of the sequence.

        Args:
            count: The number of items to skip.
        """
        return Seq(pipe(self, skip(count)))

    def sum(self: Seq[_TSupportsSum]) -> _TSupportsSum:
        """Returns the sum of the elements in the sequence."""
        return sum(self)

    def sum_by(self, projection: Callable[[_TSource], _TSupportsSum]) -> _TSupportsSum:
        """Returns the sum of the results generated by applying the
        function to each element of the sequence."""
        return pipe(self, sum_by(projection))

    def tail(self) -> Seq[_TSource]:
        """Returns a sequence that skips 1 element of the underlying
        sequence and then yields the remaining elements of the
        sequence."""
        return self.skip(1)

    def take(self, count: int) -> Seq[_TSource]:
        """Returns the first N elements of the sequence.

        Args:
            count: The number of items to take.
        """
        return Seq(pipe(self, take(count)))

    def to_list(self) -> "Block[_TSource]":
        return to_list(self)

    def dict(self) -> Iterable[_TSource]:
        """Returns a json serializable representation of the list."""

        def to_obj(value: Any) -> Any:
            attr = getattr(value, "dict", None) or getattr(value, "dict", None)
            if attr and callable(attr):
                value = attr()
            return value

        return (to_obj(value) for value in self)

    @classmethod
    def unfold(
        cls,
        generator: Callable[[_TState], Option[Tuple[_TSource, _TState]]],
        state: _TState,
    ) -> Iterable[_TSource]:
        """Returns a list that contains the elements generated by the
        given computation. The given initial state argument is passed to
        the element generator.

        Args:
            generator: A function that takes in the current state and
                returns an option tuple of the next element of the list
                and the next state value.
            state: The initial state.

        Returns:
            The result list.
        """

        return pipe(state, unfold(generator))

    def zip(self, other: Iterable[_TResult]) -> Iterable[Tuple[_TSource, _TResult]]:
        """Combines the two sequences into a list of pairs. The two
        sequences need not have equal lengths: when one sequence is
        exhausted any remaining elements in the other sequence are
        ignored.

        Args:
            other: The second input sequence.

        Returns:
            The result sequence.
        """
        return builtins.zip(self, other)

    def __iter__(self) -> Iterator[_TSource]:
        """Return iterator for sequence."""
        return builtins.iter(self._value)

    def __repr__(self) -> str:
        result = "["

        for count, x in enumerate(self):
            if count == 0:
                result += str(x)

            elif count == 100:
                result += "; ..."
                break

            else:
                result += "; " + str(x)

        return result + "]"

    def __str__(self) -> str:
        return repr(self)


class SeqGen(Iterable[_TSource]):
    """Sequence from a generator function.

    We use this to allow multiple iterations over the same sequence
    generated by a generator function."""

    def __init__(self, gen: Callable[[], Iterable[_TSource]]) -> None:
        self.gen = gen

    def __iter__(self) -> Iterator[_TSource]:
        xs = self.gen()
        return builtins.iter(xs)


def append(
    *others: Iterable[_TSource],
) -> Callable[[Iterable[_TSource]], Iterable[_TSource]]:
    """Wraps the given enumerations as a single concatenated
    enumeration."""

    def _(source: Iterable[_TSource]) -> Iterable[_TSource]:
        return concat(source, *others)

    return _


@curry_flip(1)
def choose(
    source: Iterable[_TSource], chooser: Callable[[_TSource], Option[_TResult]]
) -> Iterable[_TResult]:
    """Choose items from the sequence.

    Applies the given function to each element of the list. Returns
    the list comprised of the results x for each element where the
    function returns `Some(x)`.

    Args:
        chooser: The function to generate options from the elements.

    Returns:
        The list comprising the values selected from the chooser
        function.
    """

    def mapper(x: _TSource) -> Iterable[_TResult]:
        return chooser(x).to_seq()

    return pipe(source, collect(mapper))


@curry_flip(1)
def collect(
    source: Iterable[_TSource],
    mapping: Callable[[_TSource], Iterable[_TResult]],
) -> Iterable[_TResult]:
    def gen():
        for xs in source:
            for x in mapping(xs):
                yield x

    return SeqGen(gen)


def concat(*iterables: Iterable[_TSource]) -> Iterable[_TSource]:
    """Combines the given variable number of enumerations and/or
    enumeration-of-enumerations as a single concatenated
    enumeration.

    Args:
        iterables: The input enumeration-of-enumerations.

    Returns:
        The result sequence.
    """

    def gen():
        for it in iterables:
            for element in it:
                yield element

    return SeqGen(gen)


def delay(generator: Callable[[], Iterable[_TSource]]) -> Iterable[_TSource]:
    """Returns a sequence that is built from the given delayed
    specification of a sequence.

    The input function is evaluated each time an Iterator for the
    sequence is requested.

    Args:
        generator: The generating function for the sequence.
    """
    return SeqGen(generator)


empty: Seq[Any] = Seq()
"""The empty sequence."""


@curry_flip(1)
def filter(
    source: Iterable[_TSource], predicate: Callable[[_TSource], bool]
) -> Iterable[_TSource]:
    """Filter sequence.

    Filters the sequence to a new sequence containing only the
    elements of the sequence for which the given predicate returns
    `True`.

    Args:
        source: (curried) The input sequence to to filter.
        predicate: A function to test whether each item in the
            input sequence should be included in the output.

    Returns:
        A partially applied filter function.
    """

    return builtins.filter(predicate, source)


@curry_flip(1)
def fold(
    source: Iterable[_TSource],
    folder: Callable[[_TState, _TSource], _TState],
    state: _TState,
) -> _TState:
    """Applies a function to each element of the collection,
    threading an accumulator argument through the computation. If
    the input function is f and the elements are i0...iN then
    computes f (... (f s i0)...) iN

    Args:
        folder: A function that updates the state with each element
            from the sequence.
        state: The initial state.

    Returns:
        Partially applied fold function that takes a source sequence and
        returns the state object after the folding function is applied
        to each element of the sequence.
    """

    return functools.reduce(folder, source, state)


def fold_back(
    folder: Callable[[_TSource, _TState], _TState], source: Iterable[_TSource]
) -> Callable[[_TState], _TState]:
    """Applies a function to each element of the collection,
    starting from the end, threading an accumulator argument through
    the computation. If the input function is f and the elements are
    i0...iN then computes f i0 (... (f iN s)...)

    Args:
        folder: A function that updates the state with each element
            from the sequence.
        state: The initial state.

    Returns:
        Partially applied fold_back function.
    """

    def _fold_back(state: _TState) -> _TState:
        """Partially applied fold_back function.

        Returns:
            The state object after the folding function is applied
            to each element of the sequence.
        """
        return functools.reduce(
            lambda x, y: folder(y, x), reversed(list(source)), state
        )

    return _fold_back


def head(source: Iterable[_TSource]) -> _TSource:
    """Return the first element of the sequence.

    Args:
        source: The input sequence.

    Returns:
        The first element of the sequence.

    Raises:
        Raises `ValueError` if the source sequence is empty.
    """

    for value in source:
        return value

    raise ValueError("Sequence contains no elements")


def init_infinite(initializer: Callable[[int], _TSource]) -> Iterable[_TSource]:
    """Generates a new sequence which, when iterated, will return
    successive elements by calling the given function. The results of
    calling the function will not be saved, that is the function will be
    reapplied as necessary to regenerate the elements. The function is
    passed the index of the item being generated.

    Iteration can continue up to `sys.maxsize`.
    """

    class Infinite(Iterable[_TResult]):
        """An infinite iterable where each iterator starts counting at
        0."""

        def __init__(self, initializer: Callable[[int], _TResult]) -> None:
            self.initializer = initializer

        def __iter__(self) -> Iterator[_TResult]:
            return builtins.map(self.initializer, itertools.count(0, 1))

    return Infinite(initializer)


infinite: Iterable[int] = init_infinite(identity)
"""An infinite iterable."""


@curry_flip(1)
def iter(source: Iterable[_TSource], action: Callable[[_TSource], None]) -> None:
    """Applies the given function to each element of the collection.

    Args:
        action: A function to apply to each element of the sequence.

    Returns:
        A partially applied iter function.
    """

    for x in source:
        action(x)


def length(source: Seq[Any]) -> int:
    """Return the length of the sequence."""
    return builtins.sum(1 for _ in source)


@curry_flip(1)
def map(
    source: Iterable[_TSource], mapper: Callable[[_TSource], _TResult]
) -> Iterable[_TResult]:
    """Map source sequence.

    Builds a new collection whose elements are the results of
    applying the given function to each of the elements of the
    collection.

    Args:
        mapping: A function to transform items from the input sequence.

    Returns:
        Partially applied map function.
    """

    def gen():
        for x in source:
            yield mapper(x)

    return SeqGen(gen)


@overload
def starmap(
    mapper: Callable[[_T1, _T2], _TResult]
) -> Callable[[Iterable[Tuple[_T1, _T2]]], Iterable[_TResult]]:
    ...


@overload
def starmap(
    mapper: Callable[[_T1, _T2, _T3], _TResult]
) -> Callable[[Iterable[Tuple[_T1, _T2, _T3]]], Iterable[_TResult]]:
    ...


@overload
def starmap(
    mapper: Callable[[_T1, _T2, _T3, _T4], _TResult]
) -> Callable[[Iterable[Tuple[_T1, _T2, _T3, _T4]]], Iterable[_TResult]]:
    ...


def starmap(
    mapper: Callable[..., Any]
) -> Callable[[Iterable[Tuple[Any, ...]]], Iterable[Any]]:
    """Starmap source sequence.

    Unpack arguments grouped as tuple elements. Builds a new collection
    whose elements are the results of applying the given function to the
    unpacked arguments to each of the elements of the collection.

    Args:
        mapping: A function to transform items from the input sequence.

    Returns:
        Partially applied map function.
    """

    def mapper_(args: Tuple[Any, ...]) -> Any:
        return mapper(*args)

    return map(mapper_)


def map2(
    mapper: Callable[[_T1, _T2], _TResult]
) -> Callable[[Iterable[Tuple[_T1, _T2]]], Iterable[_TResult]]:
    return starmap(mapper)


def map3(
    mapper: Callable[[_T1, _T2, _T3], _TResult]
) -> Callable[[Iterable[Tuple[_T1, _T2, _T3]]], Iterable[_TResult]]:
    return starmap(mapper)


@curry_flip(1)
def mapi(
    source: Iterable[_TSource], mapping: Callable[[int, _TSource], _TResult]
) -> Iterable[_TResult]:
    """Map list with index.

    Builds a new collection whose elements are the results of
    applying the given function to each of the elements of the
    collection. The integer index passed to the function indicates
    the index (from 0) of element being transformed.

    Args:
        mapping: The function to transform elements and their
            indices.

    Returns:
        The list of transformed elements.
    """

    return (*itertools.starmap(mapping, builtins.enumerate(source)),)


def max(source: Iterable[_TSupportsGreaterThan]) -> _TSupportsGreaterThan:
    """Returns the greatest of all elements of the sequence,
    compared via `max()`."""

    value: _TSupportsGreaterThan = builtins.max(source)
    return value


@curry_flip(1)
def max_by(
    source: Iterable[_TSource], projection: Callable[[_TSource], _TSupportsGreaterThan]
) -> _TSupportsGreaterThan:
    return builtins.max(projection(x) for x in source)


def min(source: Iterable[_TSupportsLessThan]) -> _TSupportsLessThan:
    """Returns the smallest of all elements of the sequence,
    compared via `max()`."""

    return builtins.min(source)


@curry_flip(1)
def min_by(
    source: Iterable[_TSource], projection: Callable[[_TSource], _TSupportsLessThan]
) -> _TSupportsLessThan:
    return builtins.min(projection(x) for x in source)


def of(*args: _TSource) -> Seq[_TSource]:
    """Create sequence from iterable.

    Enables fluent dot chaining on the created sequence object.
    """
    return Seq(args)


def of_iterable(source: Iterable[_TSource]) -> Seq[_TSource]:
    """Alias to `Seq.of`."""
    return Seq(source)


of_list = of_iterable
"""Alias to `seq.of_iterable`."""


@overload
def range(stop: int) -> Iterable[int]:
    ...


@overload
def range(
    start: int,
    stop: int,
) -> Iterable[int]:
    ...


@overload
def range(start: int, stop: int, step: int) -> Iterable[int]:
    ...


def range(*args: int, **kw: int) -> Iterable[int]:
    return Seq(builtins.range(*args, **kw))


@curry_flip(1)
def scan(
    source: Iterable[_TSource],
    scanner: Callable[[_TState, _TSource], _TState],
    state: _TState,
) -> Iterable[_TState]:
    """Like fold, but computes on-demand and returns the sequence of
    intermediary and final results.

    Args:
        scanner: A function that updates the state with each element
        state: The initial state.
    """

    return itertools.accumulate(source, scanner, initial=state)


def singleton(item: _TSource) -> Seq[_TSource]:
    """Returns a sequence that yields one item only.

    Args:
        item: The input item.

    Returns:
        The result sequence of one item.
    """
    return Seq([item])


@curry_flip(1)
def skip(source: Iterable[_TSource], count: int) -> Iterable[_TSource]:
    """Returns a sequence that skips N elements of the underlying
    sequence and then yields the remaining elements of the sequence.

    Args:
        count: The number of items to skip.
    """

    def gen():
        for i, n in enumerate(source):
            if i >= count:
                yield n

    return SeqGen(gen)


def sum(source: Iterable[_TSupportsSum]) -> _TSupportsSum:
    """Returns the sum of the elements in the sequence."""
    ret = builtins.sum(source)
    return cast(_TSupportsSum, ret)


@curry_flip(1)
def sum_by(
    source: Iterable[_TSource], projection: Callable[[_TSource], _TSupportsSum]
) -> _TSupportsSum:
    """Returns the sum of the results generated by applying the
    function to each element of the sequence."""

    return sum(projection(x) for x in source)


def tail(source: Iterable[_TSource]) -> Iterable[_TSource]:
    """Returns a sequence that skips 1 element of the underlying
    sequence and then yields the remaining elements of the
    sequence."""
    proj = cast(Callable[[Iterable[_TSource]], Iterable[_TSource]], skip(1))
    return proj(source)


@curry_flip(1)
def take(source: Iterable[_TSource], count: int) -> Iterable[_TSource]:
    """Returns the first N elements of the sequence.

    Args:
        count: The number of items to take.

    Returns:
        The result sequence.
    """

    def gen():
        for i, n in enumerate(source):
            yield n

            if i == count - 1:
                break

    if count > 0:
        return SeqGen(gen)
    return Seq()


def to_list(source: Iterable[_TSource]) -> "Block[_TSource]":
    from .block import Block

    return Block.of_seq(source)


@curry_flip(1)
def unfold(
    state: _TState, generator: Callable[[_TState], Option[Tuple[_TSource, _TState]]]
) -> Iterable[_TSource]:
    """Generates a list that contains the elements generated by the
    given computation. The given initial state argument is passed to
    the element generator.

    Args:
        generator: A function that takes in the current state and
            returns an option tuple of the next element of the list
            and the next state value.

    Returns:
        A partially applied unfold function that takes the state and
        returns the result list.
    """

    while True:
        result = generator(state)
        if result.is_none():
            break

        item, state = result.value
        yield item


def zip(
    source1: Iterable[_TSource],
) -> Callable[[Iterable[_TResult]], Iterable[Tuple[_TSource, _TResult]]]:
    """Combines the two sequences into a list of pairs. The two
    sequences need not have equal lengths: when one sequence is
    exhausted any remaining elements in the other sequence are
    ignored.

    Args:
        source1: The first input sequence.

    Returns:
        Partially applied zip function.
    """

    def _zip(source2: Iterable[_TResult]) -> Iterable[Tuple[_TSource, _TResult]]:
        """Combines the two sequences into a list of pairs. The two
        sequences need not have equal lengths: when one sequence is
        exhausted any remaining elements in the other sequence are
        ignored.

        Args:
            source2: The second input sequence.

        Returns:
            The result sequence.
        """
        return builtins.zip(source1, source2)

    return _zip


__all__ = [
    "Seq",
    "append",
    "choose",
    "concat",
    "collect",
    "delay",
    "empty",
    "filter",
    "fold",
    "fold_back",
    "head",
    "iter",
    "map",
    "mapi",
    "max",
    "min",
    "min_by",
    "of",
    "of_list",
    "of_iterable",
    "range",
    "scan",
    "skip",
    "singleton",
    "sum",
    "sum_by",
    "tail",
    "take",
    "unfold",
    "zip",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from asyncio import iscoroutinefunction
from threading import RLock
from types import TracebackType
from typing import Awaitable, Callable, Optional, Type

from .error import ObjectDisposedException


class Disposable(ABC):
    """A disposable class with a context manager. Must implement the
    dispose method. Will dispose on exit."""

    @abstractmethod
    def dispose(self) -> None:
        raise NotImplementedError

    def __enter__(self) -> Disposable:
        """Enter context management."""
        return self

    def __exit__(
        self,
        exctype: Optional[Type[BaseException]],
        excinst: Optional[BaseException],
        exctb: Optional[TracebackType],
    ):
        """Exit context management."""

        self.dispose()
        return False

    @staticmethod
    def create(action: Callable[[], None]):
        """Create disposable from action. Will call action when
        disposed."""
        return AnonymousDisposable(action)


class AnonymousDisposable(Disposable):
    def __init__(self, action: Callable[[], None]):
        self._is_disposed = False
        self._action = action
        self._lock = RLock()

    def dispose(self) -> None:
        """Performs the task of cleaning up resources."""

        dispose = False
        with self._lock:
            if not self._is_disposed:
                dispose = True
                self._is_disposed = True

        if dispose:
            self._action()

    def __enter__(self) -> Disposable:
        if self._is_disposed:
            raise ObjectDisposedException()
        return self


class AsyncDisposable(ABC):
    """A disposable class with a context manager. Must implement the
    `dispose_async` method. Will dispose on exit."""

    @abstractmethod
    async def dispose_async(self) -> None:
        raise NotImplementedError

    async def __aenter__(self) -> AsyncDisposable:
        """Enter context management."""
        return self

    async def __aexit__(
        self,
        exctype: Optional[Type[BaseException]],
        excinst: Optional[BaseException],
        exctb: Optional[TracebackType],
    ) -> None:
        """Exit context management."""
        await self.dispose_async()

    @staticmethod
    def create(action: Callable[[], Awaitable[None]]) -> AsyncDisposable:
        return AsyncAnonymousDisposable(action)

    @staticmethod
    def composite(*disposables: AsyncDisposable) -> AsyncDisposable:
        return AsyncCompositeDisposable(*disposables)

    @staticmethod
    def empty() -> AsyncDisposable:
        async def anoop() -> None:
            pass

        return AsyncAnonymousDisposable(anoop)


class AsyncAnonymousDisposable(AsyncDisposable):
    def __init__(self, action: Callable[[], Awaitable[None]]) -> None:
        assert iscoroutinefunction(action)
        self._is_disposed = False
        self._action = action

    async def dispose_async(self) -> None:
        if self._is_disposed:
            return

        self._is_disposed = True
        await self._action()

    async def __aenter__(self) -> AsyncDisposable:
        if self._is_disposed:
            raise ObjectDisposedException()
        return self


class AsyncCompositeDisposable(AsyncDisposable):
    def __init__(self, *disposables: AsyncDisposable) -> None:
        self._disposables = disposables

    async def dispose_async(self) -> None:
        for disposable in self._disposables:
            await disposable.dispose_async()


__all__ = [
    "Disposable",
    "AnonymousDisposable",
    "AsyncDisposable",
    "AsyncAnonymousDisposable",
    "AsyncCompositeDisposable",
]



from typing import Any, Callable, Literal, Tuple, TypeVar, overload

from typing_extensions import Concatenate, ParamSpec

_P = ParamSpec("_P")
_A = TypeVar("_A")
_B = TypeVar("_B")
_C = TypeVar("_C")
_D = TypeVar("_D")
_E = TypeVar("_E")

_Arity = Literal[0, 1, 2, 3, 4]


def _curry(
    args: Tuple[Any, ...], arity: int, fun: Callable[..., Any]
) -> Callable[..., Any]:
    def wrapper(*args_: Any, **kw: Any) -> Any:
        if arity == 1:
            return fun(*args, *args_, **kw)
        return _curry(args + args_, arity - 1, fun)

    return wrapper


@overload
def curry(num_args: Literal[0]) -> Callable[[Callable[_P, _B]], Callable[_P, _B]]:
    ...


@overload
def curry(
    num_args: Literal[1],
) -> Callable[[Callable[Concatenate[_A, _P], _B]], Callable[[_A], Callable[_P, _B]]]:
    ...


@overload
def curry(
    num_args: Literal[2],
) -> Callable[
    [Callable[Concatenate[_A, _B, _P], _C]],
    Callable[
        [_A],
        Callable[
            [_B],
            Callable[_P, _C],
        ],
    ],
]:
    ...


@overload
def curry(
    num_args: Literal[3],
) -> Callable[
    [Callable[Concatenate[_A, _B, _C, _P], _D]],
    Callable[
        [_A],
        Callable[
            [_B],
            Callable[
                [_C],
                Callable[_P, _D],
            ],
        ],
    ],
]:
    ...


@overload
def curry(
    num_args: Literal[4],
) -> Callable[
    [Callable[Concatenate[_A, _B, _C, _D, _P], _E]],
    Callable[
        [_A],
        Callable[
            [_B],
            Callable[
                [_C],
                Callable[[_D], Callable[_P, _E]],
            ],
        ],
    ],
]:
    ...


def curry(num_args: _Arity) -> Callable[..., Any]:
    """A curry decorator.

    Makes a function curried.

    Args:
        num_args: The number of args to curry from the start of the
        function

    Example:
        >>> @curry(1)
        ... def add(a: int, b: int) -> int:
        ...    return a + b
        >>>
        >>> assert add(3)(4) == 7
    """

    def wrapper(fun: Callable[..., Any]) -> Callable[..., Any]:
        return _curry((), num_args + 1, fun)

    return wrapper


@overload
def curry_flip(
    num_args: Literal[0],
) -> Callable[[Callable[_P, _A]], Callable[_P, _A]]:
    ...


@overload
def curry_flip(
    num_args: Literal[1],
) -> Callable[[Callable[Concatenate[_A, _P], _B]], Callable[_P, Callable[[_A], _B]]]:
    ...


@overload
def curry_flip(
    num_args: Literal[2],
) -> Callable[
    [Callable[Concatenate[_A, _B, _P], _C]],
    Callable[
        _P,
        Callable[
            [_A],
            Callable[[_B], _C],
        ],
    ],
]:
    ...


@overload
def curry_flip(
    num_args: Literal[3],
) -> Callable[
    [Callable[Concatenate[_A, _B, _C, _P], _D]],
    Callable[
        _P,
        Callable[
            [_A],
            Callable[
                [_B],
                Callable[[_C], _D],
            ],
        ],
    ],
]:
    ...


@overload
def curry_flip(
    num_args: Literal[4],
) -> Callable[
    [Callable[Concatenate[_A, _B, _C, _D, _P], _E]],
    Callable[
        _P,
        Callable[
            [_A],
            Callable[
                [_B],
                Callable[[_C], Callable[[_D], _E]],
            ],
        ],
    ],
]:
    ...


def curry_flip(
    num_args: _Arity,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """A flipped curry decorator.

    Makes a function curried, but flips the curried arguments to become
    the last arguments. This is very nice when having e.g optional
    arguments after a source argument that will be piped.

    Args:
        num_args: The number of args to curry from the start of the
        function

    Example:
        >>> @curry_flip(1)
        ... def map(source: List[int], mapper: Callable[[int], int]):
        ...    return [mapper(x) for x in source]
        >>>
        >>> ys = pipe(xs, map(lambda x: x * 10))
    """

    def _wrap_fun(fun: Callable[..., Any]) -> Callable[..., Any]:
        def _wrap_args(*args: Any, **kwargs: Any) -> Callable[..., Any]:
            def _wrap_curried(*curry_args: Any) -> Any:
                return fun(*curry_args, *args, **kwargs)

            return _curry((), num_args, _wrap_curried)

        return _wrap_args if num_args else fun

    return _wrap_fun


__all__ = ["curry", "curry_flip"]



"""
Collection abstractions.
"""
from . import array, asyncseq, block, map, seq
from .array import TypedArray
from .block import Block
from .map import Map
from .seq import Seq

__all__ = [
    "array",
    "asyncseq",
    "Block",
    "block",
    "Map",
    "map",
    "Seq",
    "seq",
    "TypedArray",
]



# Attribution to original authors of this code
# --------------------------------------------
# This code has been originally been ported from the Fable project which
# was originally ported from the FSharp project.
#
# Fable:
#   https://fable.io
# - Copyright (c) Alfonso Garcia-Caro and contributors.
# - MIT License
# - https://github.com/fable-compiler/Fable/blob/nagareyama/src/fable-library/Map.fs
#
# F#
# - https://github.com/dotnet/fsharp
# - Copyright (c) Microsoft Corporation. All Rights Reserved.
# - MIT License
# - https://github.com/fsharp/fsharp/blob/master/src/fsharp/FSharp.Core/map.fs
from __future__ import annotations

from typing import (
    Any,
    Callable,
    ItemsView,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    cast,
)

from expression.core import Option, PipeMixin, SupportsLessThan, curry_flip, pipe

from . import maptree, seq
from .block import Block
from .maptree import MapTree

Key = TypeVar("Key", bound=SupportsLessThan)
Value = TypeVar("Value")
Result = TypeVar("Result")

T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")
T6 = TypeVar("T6")


class Map(Mapping[Key, Value], PipeMixin):
    """The immutable map class."""

    def __init__(self, __tree: Optional[MapTree[Key, Value]] = None) -> None:
        self._tree: MapTree[Key, Value] = __tree if __tree else maptree.empty

    def add(self, key: Key, value: Value) -> Map[Key, Value]:
        return Map(maptree.add(key, value, self._tree))

    @staticmethod
    def create(ie: Iterable[Tuple[Key, Value]]) -> Map[Key, Value]:
        return create(ie)

    def contains_key(self, key: Key) -> bool:
        return maptree.mem(key, self._tree)

    def change(
        self, key: Key, f: Callable[[Option[Value]], Option[Value]]
    ) -> Map[Key, Value]:
        return Map(maptree.change(key, f, self._tree))

    @staticmethod
    def empty() -> Map[Key, Value]:
        return Map(maptree.empty)

    def is_empty(self) -> bool:
        return maptree.is_empty(self._tree)

    def exists(self, predicate: Callable[[Key, Value], bool]) -> bool:
        return maptree.exists(predicate, self._tree)

    def filter(self, predicate: Callable[[Key, Value], bool]) -> Map[Key, Value]:
        return Map(maptree.filter(predicate, self._tree))

    def for_all(self, predicate: Callable[[Key, Value], bool]) -> bool:
        """Returns true if the given predicate returns true for all of
        the bindings in the map.

        Args:
            predicate: The function to test the input elements.

        Returns:
            True if the predicate evaluates to true for all of the
            bindings in the map.
        """
        return maptree.forall(predicate, self._tree)

    def iterate(self, f: Callable[[Key, Value], None]) -> None:
        return maptree.iter(f, self._tree)

    #     def MapRange (f:'Value->'Result) =
    #         return Map<'Key, 'Result>(comparer, maptree.map f tree)

    def fold(
        self, folder: Callable[[Result, Tuple[Key, Value]], Result], state: Result
    ) -> Result:
        return maptree.fold(folder, state, self._tree)

    def fold_back(
        self, folder: Callable[[Tuple[Key, Value], Result], Result], state: Result
    ) -> Result:
        return maptree.fold_back(folder, self._tree, state)

    def map(self, mapping: Callable[[Key, Value], Result]) -> Map[Key, Result]:
        """Builds a new collection whose elements are the results of
        applying the given function to each of the elements of the
        collection. The key passed to the function indicates the key of
        element being transformed.

        Args:
            mapping: The function to transform the key/value pairs

        Returns:
            The resulting map of keys and transformed values.
        """
        return Map(maptree.map(mapping, self._tree))

    def partition(
        self, predicate: Callable[[Key, Value], bool]
    ) -> Tuple[Map[Key, Value], Map[Key, Value]]:
        r1, r2 = maptree.partition(predicate, self._tree)
        return Map(r1), Map(r2)

    # @overload
    # def get(self, key: Key) -> Optional[Value]:
    #    ...

    # @overload
    # def get(self, key: Key, default: Value) -> Value:
    #    ...

    # def get(self, key: Key, default: Union[Value, _T]) -> Union[Value, _T]:
    #    for value in self.try_find(key):
    #        return value

    #   return default

    def items(self) -> ItemsView[Key, Value]:
        items = maptree.to_seq(self._tree)
        return ItemsView(dict(items))

    def remove(self, key: Key) -> Map[Key, Value]:
        return Map(maptree.remove(key, self._tree))

    def to_list(self) -> Block[Tuple[Key, Value]]:
        return maptree.to_list(self._tree)

    def to_seq(self) -> Iterable[Tuple[Key, Value]]:
        """Convert to sequence.

        Returns:
            Sequence of key, value tuples.
        """
        return maptree.to_seq(self._tree)

    def try_get_value(self, key: Key, value: List[Value]):
        for v in maptree.try_find(key, self._tree).to_list():
            value.append(v)
            return True
        else:
            return False

    def try_find(self, key: Key) -> Option[Value]:
        return maptree.try_find(key, self._tree)

    def try_pick(
        self, chooser: Callable[[Key, Value], Option[Result]]
    ) -> Option[Result]:
        return maptree.try_pick(chooser, self._tree)

    @staticmethod
    def of(**args: Value) -> Map[str, Value]:
        return Map(maptree.of_seq(args.items()))

    @staticmethod
    def of_block(lst: Block[Tuple[Key, Value]]) -> Map[Key, Value]:
        """Generate map from list.

        Returns:
            The new map.
        """
        return of_block((lst))

    @staticmethod
    def of_list(lst: List[Tuple[Key, Value]]) -> Map[Key, Value]:
        """Generate map from list.

        Returns:
            The new map.
        """
        return of_list((lst))

    @staticmethod
    def of_seq(sequence: Iterable[Tuple[Key, Value]]) -> Map[Key, Value]:
        """Generate map from sequence.

        Generates a new map from an iterable of key/value tuples. This
        is an alias for `Map.create`.

        Returns:
            The new map.
        """
        return of_seq(sequence)

    def __hash__(self) -> int:
        def combine_hash(x: int, y: int) -> int:
            return (x << 1) + y + 631

        res = 0
        for x, y in maptree.mk_iterator(self._tree):
            res = combine_hash(res, hash(x))
            res = combine_hash(res, hash(y))
        return res

    def __getitem__(self, k: Key) -> Value:
        return maptree.find(k, self._tree)

    def __iter__(self) -> Iterator[Key]:
        xs = maptree.mk_iterator(self._tree)
        return (k for (k, _) in xs)

    def __len__(self) -> int:
        """Return the number of bindings in the map."""
        return maptree.size(self._tree)

    def __contains__(self, o: Any) -> bool:
        return self.contains_key(o)

    def __eq__(self, o: Any) -> bool:
        if not isinstance(o, Map):
            return False

        other = cast(Map[Any, Any], o)
        iterator: Iterator[Tuple[Any, Any]] = iter(other.to_seq())

        for kv in self.to_seq():
            try:
                kv_other = next(iterator)
            except StopIteration:
                return False
            else:
                if kv != kv_other:
                    return False
        return True

    def __bool__(self) -> bool:
        return not maptree.is_empty(self._tree)

    def __str__(self) -> str:
        def to_str(item: Tuple[Key, Value]) -> str:
            key, value = item
            if isinstance(key, str):
                return f'("{key}", {value})'
            return f"({key}, {value})"

        items = pipe(self.to_seq(), seq.map(to_str))
        return f"map [{'; '.join(items)}]"

    def __repr__(self) -> str:
        return str(self)


@curry_flip(1)
def add(table: Map[Key, Value], key: Key, value: Value) -> Map[Key, Value]:
    """Add key with value to map.

    Returns a new map with the binding added to the given map. If a
    binding with the given key already exists in the input map, the
    existing binding is replaced by the new binding in the result
    map.

    Args:
        key: The input key.
        value: The input value.

    Returns:
        A partially applied add function that takes the input map and returns
        the output map.
    """

    return table.add(key, value)


@curry_flip(1)
def change(
    table: Map[Key, Value], key: Key, fn: Callable[[Option[Value]], Option[Value]]
) -> Map[Key, Value]:
    """Returns a new map with the value stored under key changed
    according to f.

    Args:
        key: The input key.
        fn: The change function.
        table: The input table.

    Returns:
        The input key.
    """

    return table.change(key, fn)


@curry_flip(1)
def contains_key(table: Map[Key, Any], key: Key) -> bool:
    return table.contains_key(key)


def count(table: Map[Any, Any]) -> int:
    """Return the number of bindings in the map."""
    return len(table)


def create(ie: Iterable[Tuple[Key, Value]]) -> Map[Key, Value]:
    return Map(maptree.of_seq(ie))


@curry_flip(1)
def find(table: Map[Key, Value], key: Key) -> Value:
    """Lookup an element in the map, raising KeyNotFoundException if no
    binding exists in the map

    Args:
        key: The key to find.
        table: The map to find the key in.

    """

    return table[key]


def is_empty(table: Map[Any, Any]) -> bool:
    """Is the map empty?

    Args:
        table: The input map.

    Returns:
        True if the map is empty.
    """
    return table.is_empty()


def iterate(action: Callable[[Key, Value], None]) -> Callable[[Map[Key, Value]], None]:
    def _iterate(table: Map[Key, Value]) -> None:
        return table.iterate(action)

    return _iterate


@curry_flip(1)
def try_pick(
    table: Map[Key, Value], chooser: Callable[[Key, Value], Option[Result]]
) -> Option[Result]:
    """Searches the map looking for the first element where the given
    function returns a Some value.

    Args:
        chooser: The function to generate options from the key/value
            pairs.
    Returns:
        Partially applied `try_pick` function that takes the input map
        and returns the first result.
    """

    return table.try_pick(chooser)


@curry_flip(1)
def pick(
    table: Map[Key, Value], chooser: Callable[[Key, Value], Option[Result]]
) -> Result:
    for res in table.try_pick(chooser):
        return res
    else:
        raise KeyError()


@curry_flip(1)
def exists(table: Map[Key, Value], predicate: Callable[[Key, Value], bool]) -> bool:
    """Returns true if the given predicate returns true for one of the
    bindings in the map.

    Args:
        predicate: The function to test the input elements.

    Returns:
        Partially applied function that takes a map table and returns
        true if the predicate returns true for one of the key/value
        pairs.
    """

    return table.exists(predicate)


@curry_flip(1)
def filter(
    table: Map[Key, Value], predicate: Callable[[Key, Value], bool]
) -> Map[Key, Value]:
    return table.filter(predicate)


@curry_flip(1)
def for_all(table: Map[Key, Value], predicate: Callable[[Key, Value], bool]) -> bool:
    return table.for_all(predicate)


@curry_flip(1)
def map(
    table: Map[Key, Value], mapping: Callable[[Key, Value], Result]
) -> Map[Key, Result]:
    return table.map(mapping)


@curry_flip(1)
def fold(
    table: Map[Key, Value],
    folder: Callable[[Result, Tuple[Key, Value]], Result],
    state: Result,
) -> Result:
    return table.fold(folder, state)


@curry_flip(1)
def fold_back(
    state: Result,
    folder: Callable[[Tuple[Key, Value], Result], Result],
    table: Map[Key, Value],
) -> Result:
    return table.fold_back(folder, state)


@curry_flip(1)
def partition(
    table: Map[Key, Value], predicate: Callable[[Key, Value], bool]
) -> Tuple[Map[Key, Value], Map[Key, Value]]:
    return table.partition(predicate)


@curry_flip(1)
def remove(table: Map[Key, Value], key: Key) -> Map[Key, Value]:
    """Removes an element from the domain of the map. No exception is
    raised if the element is not present.

    Args:
        key: The key to remove.
        table: The table to remove the key from.

    Returns:
        The resulting map.
    """

    return table.remove(key)


def of(**args: Value) -> Map[str, Value]:
    """Create map from arguments."""
    return Map(maptree.of_seq(args.items()))


def of_block(elements: Block[Tuple[Key, Value]]) -> Map[Key, Value]:
    return Map(maptree.of_list(elements))


def of_list(elements: List[Tuple[Key, Value]]) -> Map[Key, Value]:
    return Map(maptree.of_list(Block(elements)))


def of_seq(elements: Iterable[Tuple[Key, Value]]) -> Map[Key, Value]:
    return Map(maptree.of_seq(elements))


def to_list(table: Map[Key, Value]) -> Block[Tuple[Key, Value]]:
    return table.to_list()


def to_seq(table: Map[Key, Value]) -> Iterable[Tuple[Key, Value]]:
    return table.to_seq()


@curry_flip(1)
def try_find(table: Map[Key, Value], key: Key) -> Option[Value]:
    """Lookup an element in the map, returning a `Some` value if the
    element is in the domain of the map and `Nothing` if not.

    Args:
        key: The input key.

    Returns:
        A partially applied `try_find` function that takes a map
        instance and returns the result.
    """

    return table.try_find(key)


empty: Map[Any, Any] = Map.empty()

__all__ = [
    "Map",
    "add",
    "change",
    "create",
    "contains_key",
    "count",
    "empty",
    "exists",
    "filter",
    "find",
    "fold",
    "for_all",
    "is_empty",
    "iterate",
    "map",
    "of",
    "of_block",
    "of_list",
    "of_seq",
    "partition",
    "pick",
    "remove",
    "to_list",
    "to_seq",
    "try_find",
    "try_pick",
]



import builtins
import itertools
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Callable,
    Optional,
    TypeVar,
    overload,
)

from expression.core import pipe

TSource = TypeVar("TSource")
TResult = TypeVar("TResult")


class AsyncSeq(AsyncIterable[TSource]):
    def __init__(self, ai: AsyncIterable[TSource]):
        self._ai = ai

    async def map(self, mapper: Callable[[TSource], TResult]) -> AsyncIterable[TResult]:
        return pipe(self, map(mapper))

    @classmethod
    def empty(cls) -> AsyncIterable[Any]:
        return AsyncSeq(empty())

    @overload
    @classmethod
    def range(cls, stop: int) -> AsyncIterable[int]:
        ...

    @overload
    @classmethod
    def range(cls, start: int, stop: int) -> AsyncIterable[int]:
        ...

    @overload
    @classmethod
    def range(cls, start: int, stop: int, step: int) -> AsyncIterable[int]:
        ...

    @classmethod
    def range(cls, *args: Any, **kw: Any) -> AsyncIterable[int]:
        return AsyncSeq(range(*args, **kw))

    def __aiter__(self) -> AsyncIterator[TSource]:
        return self._ai.__aiter__()


def append(
    other: AsyncIterable[TSource],
) -> Callable[[AsyncIterable[TSource]], AsyncIterable[TSource]]:
    async def _append(source: AsyncIterable[TSource]) -> AsyncIterable[TSource]:
        async for value in source:
            yield value
        async for value in other:
            yield value

    return _append


async def empty() -> AsyncIterable[Any]:
    while False:
        yield


async def repeat(value: TSource, times: Optional[int] = None) -> AsyncIterable[TSource]:
    for value in itertools.repeat(value, times):  # type: ignore
        yield value


@overload
def range(stop: int) -> AsyncIterable[int]:
    ...


@overload
def range(start: int, stop: int) -> AsyncIterable[int]:
    ...


@overload
def range(start: int, stop: int, step: int) -> AsyncIterable[int]:
    ...


async def range(*args: int, **kw: int) -> AsyncIterable[int]:
    for value in builtins.range(*args, **kw):
        yield value


def filter(
    predicate: Callable[[TSource], bool]
) -> Callable[[AsyncIterable[TSource]], AsyncIterable[TSource]]:
    async def _filter(source: AsyncIterable[TSource]) -> AsyncIterable[TSource]:
        async for value in source:
            if predicate(value):
                yield value

    return _filter


def map(
    mapper: Callable[[TSource], TResult]
) -> Callable[[AsyncIterable[TSource]], AsyncIterable[TResult]]:
    async def _map(source: AsyncIterable[TSource]) -> AsyncIterable[TResult]:
        async for value in source:
            yield mapper(value)

    return _map


__all__ = [
    "AsyncSeq",
    "range",
    "filter",
    "map",
    "repeat",
]



"""Data structures that can be traversed from left to right, performing an action on each element."""
from typing import Any, Callable, TypeVar

from expression import effect
from expression.collections import Block, block, seq
from expression.core import Ok, Result, identity, pipe

_TSource = TypeVar("_TSource")
_TResult = TypeVar("_TResult")
_TError = TypeVar("_TError")


def traverse(
    fn: Callable[[_TSource], Result[_TResult, _TError]], lst: Block[_TSource]
) -> Result[Block[_TResult], _TError]:
    """Traverses a list of items.

    Threads an applicative computation though a list of items.
    """

    @effect.result[Block[_TResult], _TError]()
    def folder(head: _TSource, tail: Result[Block[_TResult], _TError]) -> Any:
        """Same as:
        >>> fn(head).bind(lambda head: tail.bind(lambda tail: Ok([head] + tail)))
        """
        h: _TResult = yield from fn(head)
        t: Block[_TResult] = yield from tail

        return Block([h]) + t

    state: Result[Block[_TResult], _TError] = Ok(block.empty)
    ret = pipe(
        state,
        seq.fold_back(folder, lst),
    )
    return ret


def sequence(lst: Block[Result[_TSource, _TError]]) -> Result[Block[_TSource], _TError]:
    """Execute a sequence of result returning commands and collect the
    sequence of their response."""

    return traverse(identity, lst)



from __future__ import annotations

import itertools
from abc import ABC
from typing import Any, Dict, Generic, Iterator, Optional, TypeVar, cast

from .pipe import PipeMixin
from .typing import GenericValidator, ModelField, SupportsValidation

_T = TypeVar("_T")


class Tag(Generic[_T]):
    """For creating tagged union cases.

    Args:
        tag: Optional tag number. If not set it will be generated
        automatically.

    """

    __match_args__ = ("tag",)

    INITIAL_TAG = 1000
    _count = itertools.count(start=INITIAL_TAG)

    def __init__(self, tag: Optional[int] = None) -> None:
        self.tag = tag or next(Tag._count)

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True

        if not isinstance(other, Tag):
            return False

        return self.tag == other.tag

    def __str__(self) -> str:
        return f"tag: {self.tag}"

    def __repr__(self) -> str:
        return f"tag: {self.tag}"


def tag(tag: Optional[int] = None) -> Tag[None]:
    """Convenience from creating tags.

    Less and simpler syntax since the type is given as an argument."""
    return Tag(tag)


class TypedTaggedUnion(
    SupportsValidation["TypedTaggedUnion[_T]"], Generic[_T], PipeMixin, ABC
):
    """A discriminated (tagged) union.

    Takes a value, and an optional tag that may be used for matching."""

    __match_args__ = ("tag", "value")

    def __init__(self, tag: Tag[_T], value: _T = None) -> None:
        self.value = value
        self.tag = tag

        if not hasattr(self.__class__, "_tags"):
            tags = {
                tag.tag: name
                for (name, tag) in self.__class__.__dict__.items()
                if isinstance(tag, Tag)
            }
            setattr(self.__class__, "_tags", tags)

        self.name = getattr(self.__class__, "_tags")[tag.tag]

    def dict(self) -> Any:
        tags: Dict[str, Any] = {
            k: v for (k, v) in self.__class__.__dict__.items() if v is self.tag
        }
        if self.value and hasattr(self.value, "dict"):
            value: Any = self.value.dict()  # type: ignore
        else:
            value = self.value
        return dict(tag=list(tags.keys())[0], value=value)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}: {self.name} {self.value}"

    def __repr__(self) -> str:
        return str(self)

    @classmethod
    def __get_validators__(cls) -> Iterator[GenericValidator[TypedTaggedUnion[_T]]]:
        def _validate(union: Any, field: ModelField) -> TypedTaggedUnion[_T]:
            if isinstance(union, TypedTaggedUnion):
                return cast(TypedTaggedUnion[_T], union)

            tags: Dict[str, Any] = {
                k: v for (k, v) in cls.__dict__.items() if k == union["tag"]
            }
            if field.sub_fields:
                sub_field = field.sub_fields[0]
                value, error = sub_field.validate(union["value"], {}, loc=union["tag"])
                if error:
                    raise ValueError(str(error))
            else:
                value = union["value"]
            return cls(tags[union["tag"]], value)

        yield _validate

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, TypedTaggedUnion):
            return False

        return self.tag == other.tag and self.value == other.value  # type: ignore


class TaggedUnion(TypedTaggedUnion[Any]):
    ...


class SingleCaseUnion(TypedTaggedUnion[_T]):
    """Single case union.

    Helper class to make single case tagged unions without having to
    declare the tag which is needed to make a tagged union. The name of
    the tag for a single case union is `VALUE` and may be used for
    matching purposes.
    """

    VALUE = Tag[_T]()

    def __init__(self, value: Any) -> None:
        setattr(self.__class__, "_tags", {SingleCaseUnion.VALUE.tag: "VALUE"})
        super().__init__(SingleCaseUnion.VALUE, value)


__all__ = ["SingleCaseUnion", "Tag", "TaggedUnion", "tag"]



# Attribution to original authors of this code
# --------------------------------------------
# This code has been originally ported from the Fable project (https://fable.io)
# Copyright (c) Alfonso García-Caro and contributors.
#
# The original code was authored by
# - Alfonso Garcia-Caro (https://github.com/alfonsogarciacaro)
# - ncave (https://github.com/ncave)
#
# You can find the original implementation here:
# - https://github.com/fable-compiler/Fable/blob/nagareyama/src/fable-library/MailboxProcessor.ts

from __future__ import annotations

import asyncio
from queue import SimpleQueue
from threading import RLock
from typing import Any, Awaitable, Callable, Generic, Optional, TypeVar

from expression.system import CancellationToken, OperationCanceledError

from .aiotools import Continuation, from_continuations, start_immediate

_Msg = TypeVar("_Msg")
_Reply = TypeVar("_Reply")


class AsyncReplyChannel(Generic[_Reply]):
    def __init__(self, fn: Callable[[_Reply], None]) -> None:
        self.fn = fn

    def reply(self, r: _Reply) -> None:
        self.fn(r)


class MailboxProcessor(Generic[_Msg]):
    def __init__(self, cancellation_token: Optional[CancellationToken]) -> None:
        self.messages: SimpleQueue[_Msg] = SimpleQueue()
        self.token = cancellation_token or CancellationToken.none()
        self.loop = asyncio.get_event_loop()
        self.lock = RLock()

        # Holds the continuation i.e the `done` callback of Async.from_continuations
        # returned by `receive`.
        self.continuation: Optional[Continuation[_Msg]] = None
        self.cancel: Optional[Continuation[OperationCanceledError]] = None

    def post(self, msg: _Msg) -> None:
        """Post a message synchronously to the mailbox processor.

        This method is not asynchronous since it's very fast to execute.
        It simply adds the message to the message queue of the mailbox
        processor and returns.

        Args:
            msg: Message to post.

        Returns:
            None
        """
        self.messages.put(msg)
        self.loop.call_soon_threadsafe(self.__process_events)

    def post_and_async_reply(
        self, build_message: Callable[[AsyncReplyChannel[_Reply]], _Msg]
    ) -> Awaitable[_Reply]:
        """Post a message asynchronously to the mailbox processor and
        wait for the reply.

        Args:
            build_message: A function that takes a reply channel
            (`AsyncReplyChannel[Reply]`) and returns a message to send
            to the mailbox processor. The message should contain the
            reply channel as e.g a tuple.

        Returns:
            The reply from mailbox processor.
        """

        result: Optional[_Reply] = None
        # This is the continuation for the `done` callback of the awaiting poster.
        continuation: Optional[Continuation[_Reply]] = None

        def check_completion() -> None:
            if result is not None and continuation is not None:
                continuation(result)

        def reply_callback(res: _Reply):
            nonlocal result
            result = res
            check_completion()

        reply_channel = AsyncReplyChannel(reply_callback)
        self.messages.put(build_message(reply_channel))
        self.__process_events()

        def callback(
            done: Continuation[_Reply],
            _: Continuation[Exception],
            __: Continuation[OperationCanceledError],
        ):
            nonlocal continuation
            continuation = done
            check_completion()

        return from_continuations(callback)

    async def receive(self) -> _Msg:
        """Receive message from mailbox.

        Returns:
            An asynchronous computation which will consume the
            first message in arrival order. No thread is blocked while
            waiting for further messages. Raises a TimeoutException if
            the timeout is exceeded.
        """

        def callback(
            done: Continuation[_Msg],
            error: Continuation[Exception],
            cancel: Continuation[OperationCanceledError],
        ):
            if self.continuation:
                raise Exception("Receive can only be called once!")

            self.continuation = done
            self.cancel = cancel

            self.__process_events()

        return await from_continuations(callback)

    def __process_events(self):
        # Cancellation of async workflows is more tricky in Python than
        # with F# so we check the cancellation token for each process.
        if self.token.is_cancellation_requested:
            self.cancel, cancel = None, self.cancel
            if cancel is not None:
                cancel(OperationCanceledError("Mailbox was cancelled"))
            return

        if self.continuation is None:
            return

        with self.lock:
            if self.messages.empty():
                return
            msg = self.messages.get()
            self.continuation, cont = None, self.continuation

            if cont is not None:  # type: ignore
                cont(msg)

    @staticmethod
    def start(
        body: Callable[[MailboxProcessor[Any]], Awaitable[None]],
        cancellation_token: Optional[CancellationToken] = None,
    ) -> MailboxProcessor[Any]:
        mbox: MailboxProcessor[Any] = MailboxProcessor(cancellation_token)
        start_immediate(body(mbox), cancellation_token)
        return mbox


__all__ = ["AsyncReplyChannel", "MailboxProcessor"]



from __future__ import annotations

import functools
from typing import Awaitable, Callable, Generic, TypeVar

from typing_extensions import ParamSpec

_TResult = TypeVar("_TResult")
_P = ParamSpec("_P")


class TailCall(Generic[_P]):
    """Returns a tail call.

    If a `tailrec` decorated function return a `TailCall` then the
    function will be called again with the new arguments provided.
    """

    def __init__(self, *args: _P.args, **kw: _P.kwargs):
        self.args = args
        self.kw = kw


TailCallResult = _TResult | TailCall[_P]


def tailrec(fn: Callable[_P, TailCallResult[_TResult, _P]]) -> Callable[_P, _TResult]:
    """Tail call recursive function decorator.

    Can be used to create tail call recursive functions that will not
    stack overflow. To recurse the function needs to return an instance
    of `TailCall` with the next arguments to be used for the next call.
    """

    def trampoline(bouncer: TailCallResult[_TResult, _P]) -> _TResult:
        while isinstance(bouncer, TailCall):
            args, kw = bouncer.args, bouncer.kw
            bouncer = fn(*args, **kw)

        return bouncer

    @functools.wraps(fn)
    def wrapper(*args: _P.args, **kw: _P.kwargs) -> _TResult:
        return trampoline(fn(*args, **kw))

    return wrapper


def tailrec_async(
    fn: Callable[_P, Awaitable[TailCallResult[_TResult, _P]]]
) -> Callable[_P, Awaitable[_TResult]]:
    """Tail call recursive async function decorator."""

    async def trampoline(bouncer: TailCallResult[_TResult, _P]) -> _TResult:
        while isinstance(bouncer, TailCall):
            args, kw = bouncer.args, bouncer.kw
            bouncer = await fn(*args, **kw)

        return bouncer

    @functools.wraps(fn)
    async def wrapper(*args: _P.args, **kw: _P.kwargs) -> _TResult:
        result = await fn(*args, **kw)
        return await trampoline(result)

    return wrapper


__all__ = ["TailCall", "TailCallResult", "tailrec", "tailrec_async"]



"""
Expression aims to be a solid, type-safe, pragmatic, and high
performance library for practical functional programming in Python 3.9+.
By pragmatic we mean that the goal of the library is to use simple
abstractions to enable you to do practical and productive functional
programming in Python.

GitHub: https://github.com/cognitedata/Expression
"""

from . import collections, core, effect
from ._version import __version__
from .core import (
    AsyncReplyChannel,
    Builder,
    Choice,
    Choice1of2,
    Choice1of3,
    Choice2,
    Choice2of2,
    Choice2of3,
    Choice3,
    Choice3of3,
    EffectError,
    Error,
    Failure,
    MailboxProcessor,
    Nothing,
    Nothing_,
    Ok,
    Option,
    Result,
    SingleCaseUnion,
    Some,
    Success,
    Tag,
    TaggedUnion,
    TailCall,
    TailCallResult,
    Try,
    compose,
    curry,
    curry_flip,
    default_arg,
    downcast,
    failwith,
    flip,
    fst,
    identity,
    is_error,
    is_none,
    is_ok,
    is_some,
    option,
    pipe,
    pipe2,
    pipe3,
    result,
    snd,
    tag,
    tailrec,
    tailrec_async,
    try_downcast,
    upcast,
)

curry_flipped = curry_flip
"""Deprecated.

Will be removed in next major version. Use curry_flip instead.
"""

__all__ = [
    "AsyncReplyChannel",
    "Builder",
    "Choice",
    "Choice1of2",
    "Choice1of3",
    "Choice2",
    "Choice2of2",
    "Choice2of3",
    "Choice3",
    "Choice3of3",
    "collections",
    "compose",
    "core",
    "curry",
    "curry_flip",
    "curry_flipped",
    "default_arg",
    "downcast",
    "effect",
    "EffectError",
    "Error",
    "Failure",
    "failwith",
    "flip",
    "fst",
    "identity",
    "is_error",
    "is_none",
    "is_ok",
    "is_some",
    "MailboxProcessor",
    "Nothing",
    "Nothing_",
    "Ok",
    "option",
    "Option",
    "pipe",
    "pipe2",
    "pipe3",
    "result",
    "Result",
    "SingleCaseUnion",
    "snd",
    "Some",
    "Success",
    "Tag",
    "TaggedUnion",
    "TailCall",
    "TailCallResult",
    "tailrec",
    "tailrec_async",
    "Try",
    "tag",
    "try_downcast",
    "upcast",
    "__version__",
]



"""
A collection of computational expression effects.
"""
from .option import OptionBuilder as option
from .result import ResultBuilder as result
from .result import TryBuilder as try_
from .seq import SeqBuilder as seq

__all__ = ["option", "result", "seq", "try_"]



from __future__ import annotations

import string
from typing import (
    Any,
    Callable,
    Generic,
    List,
    Optional,
    Tuple,
    TypeVar,
    cast,
    overload,
)

from expression.collections import Block, block
from expression.core import Error, Nothing, Ok, Option, Result, Some, curry, fst, pipe

_A = TypeVar("_A")
_B = TypeVar("_B")
_C = TypeVar("_C")
_D = TypeVar("_D")
_E = TypeVar("_E")

Remaining = Tuple[str, int]
ParseResult = Result[Tuple[_A, Remaining], str]


class Parser(Generic[_A]):
    """The Parser class."""

    __slots__ = ["_name", "_run"]

    def __init__(
        self, run: Callable[[Remaining], ParseResult[_A]], name: Optional[str] = None
    ) -> None:
        self._run = run
        self._name = name or "parser"

    def __call__(self, input: str) -> Result[_A, str]:
        """Returns result without the remaining string."""
        return self._run((input, 0)).map(fst)

    def run(self, input: Remaining) -> ParseResult[_A]:
        """Returns parser result and the remaining string."""
        return self._run(input)

    @staticmethod
    def fail(error: str) -> Parser[Any]:
        return fail(error)

    @staticmethod
    def pchar(char: str) -> Parser[str]:
        return pchar(char)

    @staticmethod
    def any_of(list_of_chars: str) -> Parser[str]:
        return any_of(list_of_chars)

    def and_then(self, p2: Parser[_B]) -> Parser[Tuple[_A, _B]]:
        return pipe(self, and_then(p2))

    def ignore_then(self, p2: Parser[_B]) -> Parser[_B]:
        return pipe(self, ignore_then(p2))

    def then_ignore(self, p2: Parser[Any]) -> Parser[_A]:
        return pipe(self, then_ignore(p2))

    def or_else(self, parser2: Parser[_A]) -> Parser[_A]:
        return or_else(self)(parser2)

    def map(self, mapper: Callable[[_A], _B]) -> Parser[_B]:
        return pipe(self, map(mapper))

    @overload
    def starmap(
        self: Parser[Tuple[_B, _C]], mapper: Callable[[_B, _C], _D]
    ) -> Parser[_D]:
        ...

    @overload
    def starmap(
        self: Parser[Tuple[_B, _C, _D]], mapper: Callable[[_B, _C, _D], _E]
    ) -> Parser[_E]:
        ...

    def starmap(
        self: Parser[Tuple[Any, ...]], mapper: Callable[..., Any]
    ) -> Parser[Any]:
        return pipe(self, starmap(mapper))

    def opt(self) -> Parser[Option[_A]]:
        return opt(self)

    def ignore(self) -> Parser[None]:
        return ignore(self)

    def bind(self, binder: Callable[[_A], Parser[_B]]) -> Parser[_B]:
        return pipe(self, bind(binder))

    def between(self, p2: Parser[_A], p3: Parser[Any]) -> Parser[_A]:
        return between(p2)(p3)(self)

    def __repr__(self) -> str:
        return self._name

    def __str__(self) -> str:
        return repr(self)


def pchar(char: str) -> Parser[str]:
    """Parse the given character."""

    def run(input: Remaining) -> ParseResult[str]:
        remaining, pos = input
        if pos >= len(remaining):
            return Error("no more input")

        first = remaining[pos]  # input[0], remaining = input[1:]
        if first == char:
            return Ok((char, (remaining, pos + 1)))
        else:
            msg = f"Expecting '{char}'. Got '{first}'"
            return Error(msg)

    return Parser(run, f"pchar('{char}')")


@curry(1)
def and_then(p2: Parser[_B], p1: Parser[_A]) -> Parser[Tuple[_A, _B]]:
    """The parser p1 .>>. p2 applies the parsers p1 and p2 in sequence
    and returns the results in a tuple.

    Args:
        p2 (Parser[_B]): Second parser.
        p1 (Parser[_A]): First parser.

    Returns:
        Parser[Tuple[_A, _B]]: Result parser.
    """

    def run(input: Remaining) -> ParseResult[Tuple[_A, _B]]:
        result1 = p1.run(input)
        match result1:
            case Error(error):
                return Error(error)
            case Ok((value1, remaining1)):
                result2 = p2.run(remaining1)

                match result2:
                    case Error(error):
                        return Error(error)
                    case Ok((value2, remaining2)):
                        return Ok(((cast(_A, value1), cast(_B, value2)), remaining2))
                    case _:
                        return Error("parser error")
            case _:
                return Error[Tuple[Tuple[_A, _B], Remaining], str]("parser error")

    return Parser(run, f"and_then({p2}, {p1}")


@curry(1)
def or_else(p1: Parser[_A], p2: Parser[_A]) -> Parser[_A]:
    def run(input: Remaining) -> ParseResult[_A]:
        result1 = p1.run(input)
        match result1:
            case Ok():
                return cast(ParseResult[_A], result1)
            case _:
                result2 = p2.run(input)
                return result2

    return Parser(run, f"or_else({p2}, {p1}")


def choice(list_of_parsers: Block[Parser[_A]]) -> Parser[_A]:
    return list_of_parsers.reduce(lambda a, b: or_else(a)(b))


def any_of(list_of_chars: str) -> Parser[str]:
    return pipe(
        list_of_chars,
        block.of_seq,
        block.map(pchar),  # convert into parsers
        choice,  # combine them
    )


parse_lowercase = any_of(string.ascii_lowercase)
parse_letters = any_of(string.ascii_letters)
parse_digit = any_of(string.digits)


@curry(1)
def map(mapper: Callable[[_A], _B], parser: Parser[_A]) -> Parser[_B]:
    def run(input: Remaining) -> ParseResult[_B]:
        # run parser with the input
        result = parser.run(input)

        # test the result for Failure/Success
        match result:
            case Ok((value, remaining)):
                # if success, return the value transformed by f
                new_value = mapper(cast(_A, value))
                return Ok[Tuple[_B, Remaining], str]((new_value, remaining))

            case Error(error):
                # if failed, return the error
                return Error[Tuple[_B, Remaining], str](error)
            case _:
                return Error[Tuple[_B, Remaining], str]("parser error")

    return Parser(run, f"map(A => B, {parser})")


@curry(1)
@overload
def starmap(
    mapper: Callable[[_A, _B], _C], parser: Parser[Tuple[_A, _B]]
) -> Parser[_C]:
    ...


@curry(1)
@overload
def starmap(
    mapper: Callable[[_A, _B, _C], _D], parser: Parser[Tuple[_A, _B, _C]]
) -> Parser[_D]:
    ...


@curry(1)
def starmap(mapper: Callable[..., Any], parser: Parser[Tuple[Any, ...]]) -> Parser[Any]:
    def mapper_(values: Tuple[Any, ...]) -> Any:
        return mapper(*values)

    return pipe(
        parser,
        map(mapper_),
    )


def preturn(x: _A) -> Parser[_A]:
    def run(input: Remaining) -> ParseResult[_A]:
        return Ok((x, input))

    return Parser(run, f"preturn({x})")


def fail(error: str) -> Parser[Any]:
    def run(input: Remaining) -> ParseResult[Any]:
        return Error[Tuple[Any, Remaining], str](error)

    return Parser(run, f'fail("{error}")')


def apply(f_p: Parser[Callable[[_A], _B]], x_p: Parser[_A]) -> Parser[_B]:
    def mapper(fx: Tuple[Callable[[_A], _B], _A]) -> _B:
        return fx[0](fx[1])

    # create a Parser containing a pair (f,x)
    return pipe(
        f_p,
        and_then(x_p),
        # map the pair by applying f to x
        map(mapper),
    )


@curry(2)
def lift2(
    fn: Callable[[_A], Callable[[_B], _C]], xP: Parser[_A], yP: Parser[_B]
) -> Parser[_C]:
    return apply(apply(preturn(fn), xP), yP)


def sequence(parser_list: Block[Parser[_A]]) -> Parser[Block[_A]]:
    # define the "cons" function, which is a two parameter function
    @curry(1)
    def cons(head: _A, tail: Block[_A]) -> Block[_A]:
        return tail.cons(head)

    # lift it to Parser World
    cons_p = lift2(cons)

    # process the list of parsers recursively
    match parser_list:
        case block.empty:
            return preturn(block.empty)
        case Block([head, *tail]):  # type: ignore
            tail_ = sequence(Block(cast(List[Parser[_A]], tail)))
            return cons_p(cast(Parser[_A], head))(tail_)


def pstring(string_input: str) -> Parser[str]:
    return pipe(
        string_input,
        block.of_seq,
        block.map(pchar),
        sequence,
        map(lambda x: "".join(x)),
    )


def parse_zero_or_more(
    parser: Parser[_A], input: Remaining
) -> Tuple[Block[_A], Remaining]:
    # run parser with the input
    first_result = parser.run(input)

    # test the result for Failure/Success
    match first_result:
        case Ok((first_value, input_after_first_parse)):
            # if parse succeeds, call recursively
            # to get the subsequent values
            subsequent_values, remaining_input = parse_zero_or_more(
                parser, input_after_first_parse
            )
            values = subsequent_values.cons(cast(_A, first_value))
            return values, remaining_input
        case _:
            return (block.empty, input)


def many(parser: Parser[_A]) -> Parser[Block[_A]]:
    def run(input: Remaining) -> ParseResult[Block[_A]]:
        # parse the input -- wrap in Success as it always succeeds
        ok = parse_zero_or_more(parser, input)
        return Ok[Tuple[Block[_A], Remaining], str](ok)

    return Parser(run, f"many({parser})")


def many1(parser: Parser[_A]) -> Parser[Block[_A]]:
    def run(input: Remaining) -> ParseResult[Block[_A]]:
        # run parser with the input
        firstResult = parser.run(input)
        # test the result for Failure/Success
        match firstResult:
            case Ok((first_value, input_after_first_parse)):
                # if first found, look for zeroOrMore now
                subsequent_values, remaining_input = parse_zero_or_more(
                    parser, input_after_first_parse
                )
                values = subsequent_values.cons(cast(_A, first_value))
                return Ok[Tuple[Block[_A], Remaining], str]((values, remaining_input))

            case Error(err):
                return Error(err)  # failed

            case _:
                return Error("parser error")

    return Parser(run, f"many({parser})")


def sep_by1(p: Parser[_A], sep: Parser[Any]) -> Parser[Block[_A]]:
    """Parses one or more occurrences of p separated by sep."""

    sep_then_p = sep.ignore_then(p)

    def mapper(p: _A, plist: Block[_A]) -> Block[_A]:
        return plist.cons(p)

    return p.and_then(many(sep_then_p)).starmap(mapper)


def sep_by(p: Parser[_A], sep: Parser[Any]) -> Parser[Block[_A]]:
    """Parses zero or more occurrences of p separated by sep."""
    return sep_by1(p, sep).or_else(preturn(block.empty))


# define parser for one or more digits
digits = many1(parse_digit)


def opt(p: Parser[_A]) -> Parser[Option[_A]]:
    nothing = cast(Option[_A], Nothing)

    def mapper(a: _A) -> Option[_A]:
        return Some(a)

    some: Parser[Option[_A]] = pipe(p, map(mapper))
    none: Parser[Option[_A]] = preturn(nothing)
    return or_else(some)(none)


@curry(1)
def then_ignore(
    p2: Parser[Any],
    p1: Parser[_A],
) -> Parser[_A]:
    """The parser p1 .>> p2 applies the parsers p1 and p2 in sequence
    and returns the result of p1.

    Args:
        p2 (Parser[_B]): Second parser.
        p1 (Parser[_A]): First parser.

    Returns:
        Parser[_A]: Result parser.
    """

    def mapper(value: Tuple[_A, Any]) -> _A:
        return value[0]

    return pipe(
        p1,
        and_then(p2),
        map(mapper),
    )


@curry(1)
def ignore_then(p2: Parser[_B], p1: Parser[Any]) -> Parser[_B]:
    """The parser p1 >>. p2 applies the parsers p1 and p2 in sequence
    and returns the result of p2.

    Args:
        p2 (Parser[_B]): Second parser
        p1 (Parser[_A]): First parser.

    Returns:
        Parser[_B]: Result parser.
    """

    def mapper(value: Tuple[Any, _B]) -> _B:
        return value[1]

    return pipe(
        p1,
        and_then(p2),
        map(mapper),
    )


def _pint() -> Parser[int]:
    # helper
    def result_to_int(sign: Option[str], digit_list: Block[str]) -> int:
        # ignore int overflow for now
        ret = pipe(
            digit_list,
            lambda digit_list: "".join(digit_list),
            int,
        )
        return ret * -1 if sign.is_some() else ret

    # map the digits to an int
    return pipe(
        "-",
        pchar,
        opt,
        and_then(digits),
        starmap(result_to_int),
    )


pint = _pint()


def _pfloat() -> Parser[float]:
    # helper
    def result_to_float(
        sd: Tuple[Option[str], Block[str]], digits2: Option[Block[str]]
    ) -> float:
        # ignore int overflow for now
        sign, digits1 = sd
        if digits2.is_some():
            digits1 = digits1 + digits2.value.cons(".")

        ret = pipe(
            digits1,
            lambda digits: "".join(digits),
            float,
        )

        return ret * -1 if sign.is_some() else ret

    # define parser for one or more digits
    digits = many1(parse_digit)

    # map the digits to an int
    ret = pipe(
        opt(pchar("-")),
        and_then(digits),
        and_then(
            opt(
                pipe(
                    pchar("."),
                    ignore_then(digits),
                )
            ),
        ),
        starmap(result_to_float),
    )
    return ret


pfloat = _pfloat()


whitespace_char = any_of(string.whitespace)
whitespace = many1(whitespace_char)


@curry(2)
def between(popen: Parser[Any], pclose: Parser[Any], p: Parser[_A]) -> Parser[_A]:
    """Parse between.

    The parser between popen pclose p applies the parsers popen, p and
    pclose in sequence. It returns the result of p.

    Args:
        popen (Parser[Any]): First parser
        pclose (Parser[Any]): Last parser
        p (Parser[_A]): Parser in between.

    Returns:
        Parser[_A]: Result of the middle (p) parser.
    """
    return pipe(
        popen,
        ignore_then(p),
        then_ignore(pclose),
    )


@curry(1)
def _starts_with(string: str, prefix: str) -> bool:
    return string.startswith(prefix)


starts_with = lift2(_starts_with)


@curry(1)
def bind(f: Callable[[_A], Parser[_B]], p: Parser[_A]) -> Parser[_B]:
    def run(input: Remaining) -> ParseResult[_B]:
        result1 = p.run(input)
        match result1:
            case Ok((value1, remaning_input)):
                p2 = f(cast(_A, value1))
                return p2.run(remaning_input)
            case Error(err):
                return Error(err)  # failed
            case _:
                return Error("parser error")

    return Parser(run, f"bind(A => Parser[B], {p}")


def ignore(p: Parser[Any]) -> Parser[None]:
    def mapper(_: Any) -> None:
        return None

    return pipe(
        p,
        map(mapper),
    )


__all__ = [
    "and_then",
    "any_of",
    "apply",
    "between",
    "bind",
    "choice",
    "fail",
    "digits",
    "map",
    "ignore",
    "ignore_then",
    "lift2",
    "many1",
    "many",
    "opt",
    "or_else",
    "parse_digit",
    "parse_letters",
    "parse_lowercase",
    "pint",
    "ParseResult",
    "Parser",
    "pchar",
    "pstring",
    "preturn",
    "sequence",
    "sep_by",
    "sep_by1",
    "starts_with",
    "then_ignore",
    "whitespace",
]



from functools import reduce
from typing import Any, Callable, TypeVar, overload

_A = TypeVar("_A")
_B = TypeVar("_B")
_C = TypeVar("_C")
_D = TypeVar("_D")
_E = TypeVar("_E")
_F = TypeVar("_F")
_G = TypeVar("_G")
_H = TypeVar("_H")
_T = TypeVar("_T")
_J = TypeVar("_J")


@overload
def compose() -> Callable[[_A], _A]:
    ...


@overload
def compose(__fn1: Callable[[_A], _B]) -> Callable[[_A], _B]:
    ...


@overload
def compose(__fn1: Callable[[_A], _B], __fn2: Callable[[_B], _C]) -> Callable[[_A], _C]:
    ...


@overload
def compose(
    __fn1: Callable[[_A], _B], __fn2: Callable[[_B], _C], __fn3: Callable[[_C], _D]
) -> Callable[[_A], _D]:
    ...


@overload
def compose(
    __fn1: Callable[[_A], _B],
    __fn2: Callable[[_B], _C],
    __fn3: Callable[[_C], _D],
    __fn4: Callable[[_D], _E],
) -> Callable[[_A], _E]:
    ...


@overload
def compose(
    __fn1: Callable[[_A], _B],
    __fn2: Callable[[_B], _C],
    __fn3: Callable[[_C], _D],
    __fn4: Callable[[_D], _E],
    __fn5: Callable[[_E], _F],
) -> Callable[[_A], _F]:
    ...


@overload
def compose(
    __fn1: Callable[[_A], _B],
    __fn2: Callable[[_B], _C],
    __fn3: Callable[[_C], _D],
    __fn4: Callable[[_D], _E],
    __fn5: Callable[[_E], _F],
    __fn6: Callable[[_F], _G],
) -> Callable[[_A], _G]:
    ...


@overload
def compose(
    __fn1: Callable[[_A], _B],
    __fn2: Callable[[_B], _C],
    __fn3: Callable[[_C], _D],
    __fn4: Callable[[_D], _E],
    __fn5: Callable[[_E], _F],
    __fn6: Callable[[_F], _G],
    __fn7: Callable[[_G], _H],
) -> Callable[[_A], _H]:
    ...


@overload
def compose(
    __fn1: Callable[[_A], _B],
    __fn2: Callable[[_B], _C],
    __fn3: Callable[[_C], _D],
    __fn4: Callable[[_D], _E],
    __fn5: Callable[[_E], _F],
    __fn6: Callable[[_F], _G],
    __fn7: Callable[[_G], _H],
    __fn8: Callable[[_H], _T],
) -> Callable[[_A], _T]:
    ...


@overload
def compose(
    __fn1: Callable[[_A], _B],
    __fn2: Callable[[_B], _C],
    __fn3: Callable[[_C], _D],
    __fn4: Callable[[_D], _E],
    __fn5: Callable[[_E], _F],
    __fn6: Callable[[_F], _G],
    __fn7: Callable[[_G], _H],
    __fn8: Callable[[_H], _T],
    __fn9: Callable[[_T], _J],
) -> Callable[[_A], _J]:
    ...


def compose(*fns: Callable[[Any], Any]) -> Callable[[Any], Any]:
    """Compose multiple functions left to right.

    Composes zero or more functions into a functional composition. The
    functions are composed left to right. A composition of zero
    functions gives back the identity function.

    >>> x = 42
    >>> f = lambda a: a * 10
    >>> g = lambda b: b + 3
    >>> h = lambda c: c / 2
    >>> compose()(x) == x
    >>> compose(f)(x) == f(x)
    >>> compose(f, g)(x) == g(f(x))
    >>> compose(f, g, h)(x) == h(g(f(x)))
    ...

    Returns:
        The composed function.
    """

    def _compose(source: Any) -> Any:
        """Return a pipeline of composed functions."""
        return reduce(lambda acc, f: f(acc), fns, source)

    return _compose


__all__ = ["compose"]



from functools import reduce
from typing import Any, Callable, TypeVar, overload

from expression.core.result import Ok, Result

_A = TypeVar("_A")
_B = TypeVar("_B")
_C = TypeVar("_C")
_D = TypeVar("_D")
_E = TypeVar("_E")
_F = TypeVar("_F")
_G = TypeVar("_G")
_TError = TypeVar("_TError")


@overload
def pipeline() -> Callable[[Any], Result[Any, Any]]:
    ...


@overload
def pipeline(
    __fn: Callable[[_A], Result[_B, _TError]]
) -> Callable[[_A], Result[_B, _TError]]:
    ...


@overload
def pipeline(
    __fn1: Callable[[_A], Result[_B, _TError]],
    __fn2: Callable[[_B], Result[_C, _TError]],
) -> Callable[[_A], Result[_C, _TError]]:
    ...


@overload
def pipeline(
    __fn1: Callable[[_A], Result[_B, _TError]],
    __fn2: Callable[[_B], Result[_C, _TError]],
    __fn3: Callable[[_C], Result[_D, _TError]],
) -> Callable[[_A], Result[_D, _TError]]:
    ...


@overload
def pipeline(
    __fn1: Callable[[_A], Result[_B, _TError]],
    __fn2: Callable[[_B], Result[_C, _TError]],
    __fn3: Callable[[_C], Result[_D, _TError]],
    __fn4: Callable[[_D], Result[_E, _TError]],
) -> Callable[[_A], Result[_E, _TError]]:
    ...


@overload
def pipeline(
    __fn1: Callable[[_A], Result[_B, _TError]],
    __fn2: Callable[[_B], Result[_C, _TError]],
    __fn3: Callable[[_C], Result[_D, _TError]],
    __fn4: Callable[[_D], Result[_E, _TError]],
    __fn5: Callable[[_E], Result[_F, _TError]],
) -> Callable[[_A], Result[_F, _TError]]:
    ...


@overload
def pipeline(
    __fn1: Callable[[_A], Result[_B, _TError]],
    __fn2: Callable[[_B], Result[_C, _TError]],
    __fn3: Callable[[_C], Result[_D, _TError]],
    __fn4: Callable[[_D], Result[_E, _TError]],
    __fn5: Callable[[_E], Result[_F, _TError]],
    __fn6: Callable[[_F], Result[_G, _TError]],
) -> Callable[[_A], Result[_G, _TError]]:
    ...


def pipeline(
    *fns: Callable[[Any], Result[Any, Any]]
) -> Callable[[Any], Result[Any, Any]]:
    """pipeline multiple result returning functions left to right.

    A pipeline kleisli (>=>) composes zero or more functions into a
    functional composition. The functions are composed left to right. A
    composition of zero functions gives back the identity function.

    >>> pipeline()(x) == Ok(x)
    >>> pipeline(f)(x) == f(x)
    >>> pipeline(f, g)(x) == f(x).bind(g)
    >>> pipeline(f, g, h)(x) == f(x).bind(g).bind(h)
    ...

    Returns:
        The composed functions.
    """

    def reducer(
        acc: Callable[[Any], Result[Any, Any]], fn: Callable[[Any], Result[Any, Any]]
    ) -> Callable[[Any], Result[Any, Any]]:
        def gn(x: Any) -> Result[Any, Any]:
            return acc(x).bind(fn)

        return gn

    return reduce(reducer, fns, Ok)


__all__ = ["pipeline"]



"""The system module.

Contains tools and utilities for dealing with (async) disposables and
cancellation tokens.
"""
from .cancellation import CancellationToken, CancellationTokenSource
from .disposable import (
    AnonymousDisposable,
    AsyncAnonymousDisposable,
    AsyncCompositeDisposable,
    AsyncDisposable,
    Disposable,
)
from .error import ObjectDisposedException, OperationCanceledError

__all__ = [
    "AsyncDisposable",
    "AsyncAnonymousDisposable",
    "AsyncCompositeDisposable",
    "CancellationToken",
    "CancellationTokenSource",
    "Disposable",
    "AnonymousDisposable",
    "ObjectDisposedException",
    "OperationCanceledError",
]



from abc import ABC
from functools import wraps
from typing import (
    Any,
    Callable,
    Generator,
    Generic,
    List,
    Optional,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import ParamSpec

from .error import EffectError

_TInner = TypeVar("_TInner")
_TOuter = TypeVar("_TOuter")
_P = ParamSpec("_P")


class Builder(Generic[_TInner, _TOuter], ABC):
    """Effect builder."""

    def bind(self, xs: _TOuter, fn: Callable[[Any], _TOuter]) -> _TOuter:
        raise NotImplementedError("Builder does not implement a bind method")

    def return_(self, x: _TInner) -> _TOuter:
        raise NotImplementedError("Builder does not implement a return method")

    def return_from(self, xs: _TOuter) -> _TOuter:
        raise NotImplementedError("Builder does not implement a return from method")

    def combine(self, xs: _TOuter, ys: _TOuter) -> _TOuter:
        """Used for combining multiple statements in the effect."""
        raise NotImplementedError("Builder does not implement a combine method")

    def zero(self) -> _TOuter:
        """Called if the effect raises StopIteration without a value,
        i.e returns None"""
        raise NotImplementedError("Builder does not implement a zero method")

    def delay(self, fn: Callable[[], _TOuter]) -> _TOuter:
        """Default implementation evaluates the given function."""
        return fn()

    def run(self, xs: _TOuter) -> _TOuter:
        """Default implementation assumes the result is already evalutated."""
        return xs

    def _send(
        self,
        gen: Generator[Any, Any, Any],
        done: List[bool],
        value: Optional[_TInner] = None,
    ) -> _TOuter:
        try:
            yielded = gen.send(value)
            return self.return_(yielded)
        except EffectError as error:
            # Effect errors (Nothing, Error, etc) short circuits the processing so we
            # set `done` to `True` here.
            done.append(True)
            return self.return_from(cast("_TOuter", error))
        except StopIteration as ex:
            done.append(True)
            # Return of a value in the generator produces StopIteration with a value
            if ex.value is not None:
                return self.return_(ex.value)
            raise
        except RuntimeError:
            done.append(True)
            raise StopIteration

    def __call__(
        self,
        fn: Callable[
            _P,
            Union[
                Generator[Optional[_TInner], _TInner, Optional[_TInner]],
                Generator[Optional[_TInner], None, Optional[_TInner]],
            ],
        ],
    ) -> Callable[_P, _TOuter]:
        """Option builder.

        Enables the use of computational expressions using coroutines.
        Thus inside the coroutine the keywords `yield` and `yield from`
        reassembles `yield` and `yield!` from F#.

        Args:
            fn: A function that contains a computational expression and
                returns either a coroutine, generator or an option.

        Returns:
            A `builder` function that can wrap coroutines into builders.
        """

        @wraps(fn)
        def wrapper(*args: _P.args, **kw: _P.kwargs) -> _TOuter:
            gen = fn(*args, **kw)
            done: List[bool] = []

            result: Optional[_TOuter] = None

            def binder(value: Any) -> _TOuter:
                ret = self._send(gen, done, value)

                # Delay every result except the first
                if result is not None:
                    return self.delay(lambda: ret)
                return ret

            try:
                result = self._send(gen, done)

                while not done:
                    cont = self.bind(result, binder)

                    # Combine every result except the first
                    if result is None:
                        result = cont
                    else:
                        result = self.combine(result, cont)
            except StopIteration:
                pass

            # If anything returns `None` (i.e raises StopIteration without a value) then
            # we expect the effect to have a zero method implemented.
            if result is None:
                result = self.zero()

            return self.run(result)

        return wrapper


__all__ = ["Builder"]



from typing import Any, Callable, Generator, Optional, TypeVar, Union

from typing_extensions import ParamSpec

from expression.core import Builder, Nothing, Option, Some, option

_TSource = TypeVar("_TSource")
_TResult = TypeVar("_TResult")
_P = ParamSpec("_P")


class OptionBuilder(Builder[_TSource, Option[Any]]):
    def bind(
        self, xs: Option[_TSource], fn: Callable[[_TSource], Option[_TResult]]
    ) -> Option[_TResult]:
        return option.bind(fn)(xs)

    def return_(self, x: _TSource) -> Option[_TSource]:
        return Some(x)

    def return_from(self, xs: Option[_TSource]) -> Option[_TSource]:
        return xs

    def combine(self, xs: Option[_TSource], ys: Option[_TSource]) -> Option[_TSource]:
        return xs.bind(lambda _: ys)

    def zero(self) -> Option[_TSource]:
        return Nothing

    def __call__(
        self,  # Ignored self parameter
        fn: Callable[
            _P,
            Union[
                Generator[Optional[_TSource], _TSource, Optional[_TSource]],
                Generator[Optional[_TSource], None, Optional[_TSource]],
            ],
        ],
    ) -> Callable[_P, Option[_TSource]]:
        return super().__call__(fn)


__all__ = ["OptionBuilder"]



from typing import Any, Callable, Generator, Optional, TypeVar, Union

from typing_extensions import ParamSpec

from expression.core import Builder, Ok, Result, pipe, result

_TSource = TypeVar("_TSource")
_TResult = TypeVar("_TResult")
_TError = TypeVar("_TError")
_P = ParamSpec("_P")


class ResultBuilder(Builder[_TSource, Result[Any, _TError]]):
    def bind(
        self,
        xs: Result[_TSource, _TError],
        fn: Callable[[_TSource], Result[_TResult, _TError]],
    ) -> Result[_TResult, _TError]:
        return pipe(xs, result.bind(fn))

    def return_(self, x: _TSource) -> Result[_TSource, _TError]:
        return Ok(x)

    def return_from(self, xs: Result[_TSource, _TError]) -> Result[_TSource, _TError]:
        return xs

    def combine(
        self, xs: Result[_TSource, _TError], ys: Result[_TSource, _TError]
    ) -> Result[_TSource, _TError]:
        return xs.bind(lambda _: ys)

    def zero(self) -> Result[_TSource, _TError]:
        raise NotImplementedError

    def __call__(
        self,  # Ignored self parameter
        fn: Callable[
            _P,
            Union[
                Generator[Optional[_TSource], _TSource, Optional[_TSource]],
                Generator[Optional[_TSource], None, Optional[_TSource]],
            ],
        ],
    ) -> Callable[_P, Result[_TSource, _TError]]:
        return super().__call__(fn)


class TryBuilder(ResultBuilder[_TSource, Exception]):
    ...


__all__ = ["ResultBuilder", "TryBuilder"]



from functools import reduce
from typing import Any, Callable, TypeVar, overload

from expression.core import Option, Some

_A = TypeVar("_A")
_B = TypeVar("_B")
_C = TypeVar("_C")
_D = TypeVar("_D")
_E = TypeVar("_E")
_F = TypeVar("_F")
_G = TypeVar("_G")
_TError = TypeVar("_TError")


@overload
def pipeline() -> Callable[[_A], Option[_A]]:
    ...


@overload
def pipeline(__fn: Callable[[_A], Option[_B]]) -> Callable[[_A], Option[_B]]:
    ...


@overload
def pipeline(
    __fn1: Callable[[_A], Option[_B]], __fn2: Callable[[_B], Option[_C]]
) -> Callable[[_A], Option[_C]]:
    ...


@overload
def pipeline(
    __fn1: Callable[[_A], Option[_B]],
    __fn2: Callable[[_B], Option[_C]],
    __fn3: Callable[[_C], Option[_D]],
) -> Callable[[_A], Option[_D]]:
    ...


@overload
def pipeline(
    __fn1: Callable[[_A], Option[_B]],
    __fn2: Callable[[_B], Option[_C]],
    __fn3: Callable[[_C], Option[_D]],
    __fn4: Callable[[_D], Option[_E]],
) -> Callable[[_A], Option[_E]]:
    ...


@overload
def pipeline(
    __fn1: Callable[[_A], Option[_B]],
    __fn2: Callable[[_B], Option[_C]],
    __fn3: Callable[[_C], Option[_D]],
    __fn4: Callable[[_D], Option[_E]],
    __fn5: Callable[[_E], Option[_F]],
) -> Callable[[_A], Option[_F]]:
    ...


@overload
def pipeline(
    __fn1: Callable[[_A], Option[_B]],
    __fn2: Callable[[_B], Option[_C]],
    __fn3: Callable[[_C], Option[_D]],
    __fn4: Callable[[_D], Option[_E]],
    __fn5: Callable[[_E], Option[_F]],
    __fn6: Callable[[_F], Option[_G]],
) -> Callable[[_A], Option[_G]]:
    ...


def pipeline(*fns: Callable[[Any], Option[Any]]) -> Callable[[Any], Option[Any]]:
    """pipeline multiple option returning functions left to right.

    A pipeline kleisli (>=>) composes zero or more functions into a
    functional composition. The functions are composed left to right. A
    composition of zero functions gives back the identity function.

    >>> pipeline()(x) == Some(x)
    >>> pipeline(f)(x) == f(x)
    >>> pipeline(f, g)(x) == f(x).bind(g)
    >>> pipeline(f, g, h)(x) == f(x).bind(g).bind(h)
    ...

    Returns:
        The composed functions.
    """

    def reducer(
        acc: Callable[[Any], Option[Any]], fn: Callable[[Any], Option[Any]]
    ) -> Callable[[Any], Option[Any]]:
        def gn(x: Any) -> Option[Any]:
            return acc(x).bind(fn)

        return gn

    return reduce(reducer, fns, Some)


__all__ = ["pipeline"]



from __future__ import annotations

from threading import RLock
from typing import Callable, Dict, List, Optional

from .disposable import Disposable
from .error import ObjectDisposedException


class CancellationToken:
    """A `CancellationToken` enables cooperative cancellation between
    threads, thread pool work items, or Task objects. You create a
    cancellation token by instantiating a `CancellationTokenSource`
    object, which manages cancellation tokens retrieved from its
    CancellationTokenSource.Token property. You then pass the
    cancellation token to any number of threads, tasks, or operations
    that should receive notice of cancellation. The token cannot be used
    to initiate cancellation. When the owning object calls
    `CancellationTokenSource.cancel`, the `is_cancellation_requested`
    property on every copy of the cancellation token is set to true. The
    objects that receive the notification can respond in whatever manner
    is appropriate."""

    def __init__(
        self, cancelled: bool = True, source: Optional[CancellationTokenSource] = None
    ) -> None:
        """Should not be used directly. Create cancellation tokens using
        the `CancellationTokenSource` instead."""

        self._cancelled = cancelled
        self._source = (
            CancellationTokenSource.cancelled_source() if source is None else source
        )

    @property
    def is_cancellation_requested(self) -> bool:
        return not self._cancelled and self._source.is_cancellation_requested

    @property
    def can_be_canceled(self) -> bool:
        return not self._cancelled

    def throw_if_cancellation_requested(self):
        if self.is_cancellation_requested:
            raise ObjectDisposedException()

    def register(self, callback: Callable[[], None]) -> Disposable:
        return self._source.register_internal(callback)

    @staticmethod
    def none():
        return CancellationToken(True, None)


class CancellationTokenSource(Disposable):
    def __init__(self):
        self._is_disposed = False
        self._lock = RLock()
        self._listeners: Dict[int, Callable[[], None]] = dict()
        self._id = 0

    @property
    def token(self) -> CancellationToken:
        return CancellationToken(False, self)

    @property
    def is_cancellation_requested(self) -> bool:
        return self._is_disposed

    def cancel(self) -> None:
        self.dispose()

    def dispose(self) -> None:
        """Performs the task of cleaning up resources."""

        listeners: List[Callable[[], None]] = []
        with self._lock:
            if not self._is_disposed:
                self._is_disposed = True
                listeners.extend(self._listeners.values())

        for listener in listeners:
            try:
                listener()
            except Exception:
                pass

    def register_internal(self, callback: Callable[[], None]) -> Disposable:
        if self._is_disposed:
            raise ObjectDisposedException()

        with self._lock:
            current = self._id
            self._listeners[current] = callback
            self._id += 1

        def dispose():
            with self._lock:
                del self._listeners[current]

        return Disposable.create(dispose)

    def __enter__(self) -> Disposable:
        if self._is_disposed:
            raise ObjectDisposedException()
        return self

    @staticmethod
    def cancelled_source() -> CancellationTokenSource:
        source = CancellationTokenSource()
        source.cancel()
        return source



from typing import NoReturn


class EffectError(Exception):
    """An error that will exit any computational expression.

    We use this to detect if sub-generators causes an exit, since
    yielding nothing will be silently ignored.
    """


def failwith(message: str) -> NoReturn:
    """Raise exception with the given message string."""
    raise Exception(message)


__all__ = ["EffectError", "failwith"]



from __future__ import annotations

from abc import abstractmethod
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Protocol,
    Tuple,
    Type,
    TypeVar,
    cast,
    get_origin,
)

_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)

_Base = TypeVar("_Base")
_Derived = TypeVar("_Derived")


class SupportsLessThan(Protocol):
    @abstractmethod
    def __lt__(self, __other: Any) -> bool:
        raise NotImplementedError


class SupportsSum(Protocol):
    @abstractmethod
    def __radd__(self, __other: Any) -> Any:
        raise NotImplementedError

    def __add__(self, __other: Any) -> Any:
        raise NotImplementedError


class SupportsGreaterThan(Protocol):
    @abstractmethod
    def __gt__(self, __other: Any) -> bool:
        raise NotImplementedError


class SupportsMatch(Protocol[_T_co]):
    """Pattern matching protocol."""

    @abstractmethod
    def __match__(self, pattern: Any) -> Iterable[_T_co]:
        """Match pattern with value.

        Return a singleton iterable item (e.g `[ value ]`) if pattern
        matches value, else an empty iterable (e.g. `[]`)."""

        raise NotImplementedError


class ModelField:
    """Type mock to avoid taking a hard dependency on pydantic."""

    sub_fields: List[ModelField]

    def validate(self, value: Any, values: Dict[str, str], loc: str) -> Tuple[Any, Any]:
        ...


Validator = Callable[[Any], _T]
GenericValidator = Callable[[Any, ModelField], _T]


class SupportsValidation(Protocol[_T_co]):
    """A type that implements __get_validators__ to be used with
    pydantic."""

    @classmethod
    def __get_validators__(
        cls,
    ) -> Iterator[Validator[_T_co] | GenericValidator[_T_co]]:
        """Yield an iterator of validators."""
        ...


def upcast(type: Type[_Base], expr: _Base) -> _Base:
    """Upcast expression from a `Derived` to `Base`.

    Note: F# `:>` or `upcast`.
    """
    return expr


def downcast(type: Type[_Derived], expr: Any) -> _Derived:
    """Downcast expression `Derived` to `Base`

    Checks at compile time that the type of expression Base is a
    supertype of Derived, and checks at runtime that Base is in fact an
    instance of Derived.

    Note: F# `:?>` or `downcast`.
    """
    assert isinstance(
        expr, type
    ), f"The type of expression {expr} is not a supertype of {type}"
    return cast(_Derived, expr)


def try_downcast(type_: Type[_Derived], expr: Any) -> Optional[_Derived]:
    """Downcast expression `Base` to `Derived`.

    Check that the `Derived` type is a supertype of `Base`.

    NOTE: Supports generic types.

    Returns:
        None if `Derived` is not a supertype of `Base`.
    """
    origin: Optional[Type[_Derived]] = get_origin(type_) or type_
    if origin is not None and isinstance(expr, origin):
        derived = cast(_Derived, expr)
        return derived

    return None


__all__ = [
    "downcast",
    "upcast",
    "try_downcast",
    "GenericValidator",
    "SupportsLessThan",
    "SupportsSum",
    "SupportsGreaterThan",
    "SupportsValidation",
    "SupportsMatch",
    "Validator",
    "GenericValidator",
]



"""
Core abstractions such as pipes, options and results.
"""

from . import aiotools, option, result
from .builder import Builder
from .choice import (
    Choice,
    Choice1of2,
    Choice1of3,
    Choice2,
    Choice2of2,
    Choice2of3,
    Choice3,
    Choice3of3,
)
from .compose import compose
from .curry import curry, curry_flip
from .error import EffectError, failwith
from .fn import TailCall, TailCallResult, tailrec, tailrec_async
from .mailbox import AsyncReplyChannel, MailboxProcessor
from .misc import flip, fst, identity, snd
from .option import Nothing, Nothing_, Option, Some, default_arg, is_none, is_some
from .pipe import PipeMixin, pipe, pipe2, pipe3
from .result import Error, Ok, Result, is_error, is_ok
from .try_ import Failure, Success, Try
from .typing import (
    SupportsGreaterThan,
    SupportsLessThan,
    SupportsMatch,
    SupportsSum,
    downcast,
    try_downcast,
    upcast,
)
from .union import SingleCaseUnion, Tag, TaggedUnion, tag

__all__ = [
    "aiotools",
    "AsyncReplyChannel",
    "Builder",
    "Choice",
    "Choice2",
    "Choice3",
    "Choice1of2",
    "Choice2of2",
    "Choice1of3",
    "Choice2of3",
    "Choice3of3",
    "compose",
    "curry",
    "curry_flip",
    "default_arg",
    "downcast",
    "EffectError",
    "Error",
    "Failure",
    "failwith",
    "flip",
    "fst",
    "identity",
    "is_error",
    "is_none",
    "is_ok",
    "is_some",
    "MailboxProcessor",
    "Nothing",
    "Nothing_",
    "Ok",
    "Option",
    "option",
    "pipe",
    "pipe2",
    "pipe3",
    "PipeMixin",
    "result",
    "Result",
    "SingleCaseUnion",
    "snd",
    "Some",
    "Success",
    "SupportsLessThan",
    "SupportsGreaterThan",
    "SupportsMatch",
    "SupportsSum",
    "Tag",
    "TaggedUnion",
    "TailCall",
    "TailCallResult",
    "tailrec",
    "tailrec_async",
    "Try",
    "tag",
    "try_downcast",
    "upcast",
]
